import os

import click
import requests
from requests.auth import HTTPBasicAuth


def load_environment_auth_vars() -> tuple:
    """Attempts to load authentication information from environment variables if none is given, looks for a username
    under "FHIR_USER", password under "FHIR_PW" and the token under "FHIR_TOKEN"

    Returns:
        Tuple containing username, password and token if they were found or None if they are not present in the env vars

    """
    username = os.getenv("FHIR_USER", None)
    password = os.getenv("FHIR_PW", None)
    token = os.getenv("FHIR_TOKEN", None)

    return username, password, token


def generate_auth(username: str = None, password: str = None, token: str = None,
                  load_env: bool = False) -> requests.auth.AuthBase:
    """Generate authentication for the request to be sent to server. Either based on a given bearer token or using basic
    auth with username and password.

    Args:
      username: username for basic auth
      password: password for basic auth
      token: token to be used for bearer auth
      load_env: whether to attempt to load environment variables for auth

    Returns:
        requests auth object to use in an API call
    """
    if (not username and not password) and not token:

        if load_env:
            click.echo("No authentication given. Attempting authentication via environment variables")
            username, password, token = load_environment_auth_vars()

        if (not username and not password) and not token:
            raise ValueError("No authentication information given.")

    if (username and password) and token:
        raise ValueError("Conflicting authentication information both token and username:password set.")

    if username and not password:
        raise ValueError(f"Missing password for user: {username}")

    if username and password:
        return HTTPBasicAuth(username=username, password=password)
    elif token:
        return BearerAuth(token=token)

    else:
        raise ValueError("No authentication information given")


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r

    def __repr__(self):
        return f"BearerAuth(token={self.token[:16]}...)"




import os
import pytest
from fhir.resources import FHIRAbstractModel
from unittest import mock

from requests.auth import HTTPBasicAuth

from fhir_kindling import FhirServer, FHIRQuery
from dotenv import load_dotenv, find_dotenv
from fhir.resources.organization import Organization
from fhir.resources.address import Address
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.patient import Patient
from fhir_kindling.generators import PatientGenerator


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def oidc_server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def fhir_server(api_url):
    print(api_url)
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def org_bundle(api_url):
    bundle = Bundle.construct()
    bundle.type = "transaction"

    entries = []

    for i in range(10):
        org = Organization.construct()
        org.name = f"Test Create Org No. {i}"
        address = Address.construct()
        address.country = "Germany"
        org.address = [address]

        entry = BundleEntry.construct()

        entry.request = BundleEntryRequest(
            **{
                "method": "POST",
                "url": org.get_resource_type()
            }
        )

        entry.resource = org.dict()
        entries.append(entry)

    bundle.entry = entries
    return bundle


def test_server_api_url_validation(api_url):
    server = FhirServer(api_address=api_url)

    assert server
    malformed_url_1 = "hello-world"

    with pytest.raises(ValueError):
        server = FhirServer(api_address=malformed_url_1)

    malformed_url_2 = "htp://hapi.fhir.org/baseR4"
    with pytest.raises(ValueError):
        server = FhirServer(api_address=malformed_url_2)

    server = FhirServer(api_address="http://blaze:8080/fhir")
    assert server

    server = FhirServer(api_address="http://bla_ze:8080/fhir")
    assert server

    server = FhirServer(api_address="http://bla-ze:8080/fhir")
    assert server

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla ze:8080/fhir")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla_ze:8080/ fhir")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla_ze:8x80/fhir")


def test_server_capabilities(fhir_server):
    capabilities = fhir_server.capabilities
    assert capabilities


def test_server_rest_resources(fhir_server: FhirServer):
    resources = fhir_server.rest_resources
    assert resources
    assert len(resources) > 1


def test_server_summary(fhir_server: FhirServer):
    summary = fhir_server.summary()
    assert summary
    print(summary)


def test_fhir_server_from_env():
    load_dotenv(find_dotenv())

    client_id = os.getenv("CLIENT_ID")
    client_secret = os.getenv("CLIENT_SECRET")
    oidc_provider_url = os.getenv("OIDC_PROVIDER_URL")

    with mock.patch.dict(os.environ, {"FHIR_API_URL": "http://test.fhir.org/r4"}):
        server = FhirServer.from_env(no_auth=True)
        assert server
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_USER": "",
                "FHIR_PW": "",
                "FHIR_TOKEN": "",
                "CLIENT_ID": "",
                "CLIENT_SECRET": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # user and no password in env
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_USER": "user",
                "FHIR_PW": "",
                "FHIR_TOKEN": "",
                "CLIENT_ID": "",
                "CLIENT_SECRET": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # correct basic auth config
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_USER": "user",
                "FHIR_PW": "password",
                "FHIR_TOKEN": "",
                "CLIENT_ID": "",
                "CLIENT_SECRET": "",
            }):
        server = FhirServer.from_env()
        assert server.auth

    # token auth
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_TOKEN": "token",
                "CLIENT_ID": "",
                "CLIENT_SECRET": "",
                "FHIR_USER": "",
                "FHIR_PW": "",
            }
    ):
        server = FhirServer.from_env()
        assert server.auth

    # conflicting auth token and user/password
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_USER": "user",
                "FHIR_PW": "password",
                "FHIR_TOKEN": "token",
                "CLIENT_ID": "",
                "CLIENT_SECRET": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # conflicting auth user and client_id/secret
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "FHIR_USER": "user",
                "FHIR_PW": "password",
                "CLIENT_ID": "token"
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # conflicting auth token and client_id/secret
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "CLIENT_ID": "token",
                "FHIR_TOKEN": "token"
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # missing client secret & provider url
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "CLIENT_ID": "token",
                "CLIENT_SECRET": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    # missing provider url
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "CLIENT_ID": "token",
                "CLIENT_SECRET": "token",
                "OIDC_PROVIDER_URL": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()
    # missing secret
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "CLIENT_ID": "token",
                "OIDC_PROVIDER_URL": "https://test.fhir.org/r4",
                "CLIENT_SECRET": "",
            }):
        with pytest.raises(EnvironmentError):
            server = FhirServer.from_env()

    print(client_id, client_secret, oidc_provider_url)
    with mock.patch.dict(
            os.environ,
            {
                "FHIR_API_URL": "http://test.fhir.org/r4",
                "CLIENT_ID": client_id,
                "CLIENT_SECRET": client_secret,
                "OIDC_PROVIDER_URL": oidc_provider_url,
                "FHIR_USER": "",
                "FHIR_PW": "",
                "FHIR_TOKEN": ""
            }):
        server = FhirServer.from_env()


def test_upload_single_resource(fhir_server: FhirServer):
    from fhir.resources.organization import Organization
    from fhir.resources.address import Address

    org = Organization.construct()
    org.name = "Test Create Org"
    address = Address.construct()
    address.country = "Germany"
    org.address = [address]

    Organization.element_properties()

    response = fhir_server.add(org)

    response2 = fhir_server.add(org.dict())

    with pytest.raises(ValueError):
        fhir_server.add({"dshadk": "sjdhka"})

    assert response


def test_server_add_all(fhir_server: FhirServer):
    generator = PatientGenerator(n=100)
    patients = generator.generate()
    add_response = fhir_server.add_all(patients)

    assert add_response


def test_query_all(fhir_server: FhirServer):
    response = fhir_server.query(Patient, output_format="dict").all()
    print(response)
    assert response.response


def test_query_with_string_resource(fhir_server: FhirServer):
    auth = fhir_server.auth
    query = FHIRQuery(fhir_server.api_address, "Patient", auth=auth)

    response = query.all()

    assert isinstance(query.resource, FHIRAbstractModel)

    assert response


def test_query_with_limit(fhir_server: FhirServer):
    response = fhir_server.query(Patient, output_format="dict").limit(2)

    assert response.response["entry"]

    assert len(response.response["entry"]) == 2


def test_query_raw_string(fhir_server: FhirServer):
    query_string = "/Patient?"
    query = fhir_server.raw_query(query_string=query_string, output_format="dict")

    assert isinstance(query, FHIRQuery)

    assert query.resource.get_resource_type() == "Patient"

    assert query.all().response["entry"]


def test_add_bundle(fhir_server: FhirServer, org_bundle):
    response = fhir_server.add_bundle(org_bundle)
    assert response
    print(response)


def test_delete(fhir_server: FhirServer):
    # create 100 patients
    generator = PatientGenerator(n=100)
    patients = generator.generate()
    add_response = fhir_server.add_all(patients)
    print(add_response.references)

    delete_response = fhir_server.delete(references=add_response.references[:50])
    print(delete_response)

    delete_resource = fhir_server.delete(resources=add_response.resources[50:75])
    print(delete_resource)
    assert delete_resource
    delete_resource = fhir_server.delete(resources=[r.dict() for r in add_response.resources[75:]])
    assert delete_resource
    print(delete_resource)

    with pytest.raises(ValueError):
        fhir_server.delete(resources=["asd"], references=["asd"])
    with pytest.raises(ValueError):
        fhir_server.delete(resources=["asd"], query=["asd"])

    with pytest.raises(ValueError):
        fhir_server.delete(references=["asd"], query=["asd"])


def test_update(fhir_server: FhirServer):
    # create 100 patients
    generator = PatientGenerator(n=2)
    patients = generator.generate()
    add_response = fhir_server.add_all(patients)
    print(add_response.references)

    # update the first patient
    patient = add_response.resources[0]
    patient.name[0].family = "Test"
    update_response = fhir_server.update([patient])
    print(update_response)

    patient.name[0].family = "Test2"
    update_response = fhir_server.update([patient.dict()])

    # no resource type in dict
    with pytest.raises(ValueError):
        update_response = fhir_server.update([{"dsada": "asdas"}])

    # invalid resource type
    with pytest.raises(ValueError):
        update_response = fhir_server.update(["sdjhadk"])


# def test_transfer(fhir_server: FhirServer):
#     origin_server = FhirServer(api_address="https://mii-agiop-cord.life.uni-leipzig.de/fhir")
#     query = origin_server.query("Condition").all()
#     print(query.resources)

def test_custom_headers():
    headers = {"X-Custom-Header": "Test", "X-Custom-Header2": "Test2"}
    server = FhirServer(api_address="https://fhir.test/fhir", headers=headers)

    assert server.session.headers["X-Custom-Header"] == "Test"
    assert server.session.headers["X-Custom-Header2"] == "Test2"
    assert server.session.headers["Content-Type"] == "application/fhir+json"


def test_custom_auth():
    auth = HTTPBasicAuth(username="test", password="test")
    server = FhirServer(api_address="https://fhir.test/fhir", auth=auth)

    assert server.session.auth == auth

    with pytest.raises(ValueError):
        server = FhirServer(api_address="https://fhir.test/fhir", auth=auth, username="test")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="https://fhir.test/fhir", auth=auth, token="test")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="https://fhir.test/fhir", auth=auth, client_id="test")


def test_server_repr_str(fhir_server: FhirServer):
    print(fhir_server)
    assert fhir_server.__repr__() == fhir_server.__str__()



import os

import pytest
from dotenv import load_dotenv, find_dotenv
from pydantic import ValidationError

from fhir_kindling import FhirServer
from fhir_kindling.generators.resource_generator import ResourceGenerator, GeneratorParameters, FieldValue
from fhir.resources.condition import Condition
from fhir.resources.patient import Patient
from fhir.resources.organization import Organization
from fhir.resources.observation import Observation
from fhir.resources.fhirtypes import ObservationType
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.immunization import Immunization
from fhir.resources.coding import Coding
from fhir.resources.reference import Reference
from fhir_kindling.generators.patient import PatientGenerator
from fhir_kindling.generators.field_generator import FieldGenerator
from fhir_kindling.generators.dataset import DatasetGenerator
from pprint import pp, pprint
import datetime
import pendulum


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def covid_code():
    covid_code = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="RA01.0",
                display="COVID-19, virus identified"
            )
        ],
        text="COVID-19"
    )
    return covid_code


@pytest.fixture
def vaccination_code():
    covid_code = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="XM0GQ8",
                display="COVID-19 vaccine, RNA based"
            )
        ],
        text="COVID vaccination"
    )
    return covid_code


@pytest.fixture
def covid_params(covid_code):
    # generate patients with ids and turn into an iterator to call as a reference
    patients = PatientGenerator(n=100, generate_ids=True).generate(display=False)
    patient_ids = iter([patient.id for patient in patients])
    params = GeneratorParameters(
        count=100,
        field_values=[
            FieldValue(field="code", value=covid_code),
        ],
        field_generators=[
            FieldGenerator(
                field="subject",
                generator_function=lambda: {"reference": f"Patient/{next(patient_ids)}"}
            )
        ]
    )

    return params, patients, patient_ids


def test_patient_generator():
    patient_generator = PatientGenerator(n=100)
    patients = patient_generator.generate()
    assert len(patients) == 100
    assert isinstance(patients[0], Patient)


    generator = PatientGenerator(n=10, age_range=(18, 60))
    patients = generator.generate()

    with pytest.raises(ValueError):
        generator = PatientGenerator(n=10, age_range=("hello", "world"))
        generator.generate()

    assert len(patients) == 10

    organization_reference = Reference(
        reference="Organization/1",
    )

    generator = PatientGenerator(n=10, age_range=(18, 60), organisation=organization_reference)

    print(generator)
    patients = generator.generate()
    assert len(patients) == 10


def test_generator_field_parameters():
    params = FieldGenerator(field="birthdate", choices=["2018-01-01", "2018-01-02"],
                            choice_probabilities=[0.5, 0.5])
    with pytest.raises(ValidationError):
        params = FieldGenerator(field="birthdate", choices=["2018-01-01", "2018-01-02"],
                                choice_probabilities=[0.5, 0.3, 0.2])

    with pytest.raises(ValidationError):
        params = FieldGenerator(field="birthdate", choices=["2018-01-01", "2018-01-02"],
                                choice_probabilities=[0.5, 0.6])

    with pytest.raises(ValidationError):
        params = FieldGenerator(field="birthdate", choices=["2018-01-01", "2018-01-02"],
                                choice_probabilities=[0.5, 0.5], generator_function=lambda x: "2018-01-01")

    with pytest.raises(ValidationError):
        params = FieldGenerator(field="birthdate")


def test_field_generator():
    choices = ["2018-01-01", "2018-01-02"]
    params = FieldGenerator(field="birthdate", choices=choices, choice_probabilities=[0.5, 0.5])
    value = params.generate()
    assert value in choices

    params = FieldGenerator(field="birthdate", choices=choices)
    value = params.generate()
    assert value in choices

    params = FieldGenerator(field="birthdate", choices=choices, choice_probabilities=[1.0, 0.0])
    for i in range(100):
        value = params.generate()
        assert value == choices[0]

    params = FieldGenerator(field="birthdate", generator_function=lambda: "2018-01-01")
    value = params.generate()
    assert value == "2018-01-01"


def test_generator_with_parameters(covid_params):
    params, patients, patient_ids = covid_params

    generator = ResourceGenerator("Condition", generator_parameters=params)

    resources = generator.generate()

    assert len(resources) == 100


def test_resource_generator(covid_code):
    patient_generator = PatientGenerator(n=100, generate_ids=True)
    patients, references = patient_generator.generate(references=True)

    references_iter = iter(references)

    params = GeneratorParameters(
        count=100,
        field_values=[

        ],
        field_generators=[
            # user patient reference iterator to generate subject field
            FieldGenerator(field="subject", generator_function=lambda: next(references_iter)),
        ]
    )

    generator = ResourceGenerator("Condition", generator_parameters=params)
    data = generator.generate()

    assert len(data) == 100

    params = GeneratorParameters(
        count=100,
        field_values=[
            FieldValue(field="subject", value=references),
        ],
    )

    generator = ResourceGenerator("Condition", generator_parameters=params)
    data = generator.generate()
    assert len(data) == 100

    # too few resources in field value
    with pytest.raises(ValueError):
        params = GeneratorParameters(
            count=200,
            field_values=[
                FieldValue(field="subject", value=references),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()

    # duplicate field values should raise error
    with pytest.raises(ValueError):
        patient_2, reference_2 = patient_generator.generate(references=True)
        params = GeneratorParameters(
            count=100,
            field_values=[
                FieldValue(field="subject", value=references),
                FieldValue(field="subject", value=reference_2),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()

    references_iter_2 = iter(references)

    # generator and value targeting the same field should raise error
    with pytest.raises(ValueError):
        patient_2, reference_2 = patient_generator.generate(references=True)
        params = GeneratorParameters(
            count=100,
            field_values=[
                FieldValue(field="subject", value=references),
                FieldGenerator(field="subject", generator_function=lambda: next(references_iter_2)),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()


def test_generate_covid_dataset(vaccination_code, covid_code, server):
    count = 100
    dataset_generator = DatasetGenerator("Patient", n=count)

    covid_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=covid_code),
        ]
    )

    covid_generator = ResourceGenerator("Condition", generator_parameters=covid_params)
    # add covid conditions to patients
    dataset_generator.add_resource(covid_generator, name="covid")

    patients, patient_ids = PatientGenerator(n=count, generate_ids=True).generate(references=True)

    vaccination_date_generator = FieldGenerator(
        field="occurrenceDateTime",
        generator_function=lambda: pendulum.now().to_date_string()
    )

    first_vax_params = GeneratorParameters(
        field_values=[
            FieldValue(field="vaccineCode", value=vaccination_code),
            FieldValue(field="status", value="completed"),
        ],
        field_generators=[
            vaccination_date_generator
        ]
    )
    vaccination_generator = ResourceGenerator("Immunization", generator_parameters=first_vax_params)
    print(vaccination_generator)
    dataset_generator.add_resource(vaccination_generator, name="first_vaccination", likelihood=0.8)

    dataset = dataset_generator.generate(ids=True)
    print(dataset)
    # pprint(result.dict(exclude_none=True))

    dataset.upload(server)



import random

from pydantic import BaseModel, validator, root_validator
from typing import List, Callable, Any, Optional


class FieldGenerator(BaseModel):
    field: str
    choices: Optional[List[Any]] = None
    choice_probabilities: Optional[List[float]] = None
    generator_function: Callable[[], Any] = None

    @validator("choice_probabilities", always=True)
    def check_probability_sum(cls, v):
        if v:
            if sum(v) != 1:
                raise ValueError("Probabilities must sum to 1")
            return v
        return None

    @root_validator
    def check_choices_and_probabilities(cls, values):
        if values.get("choices"):
            if values.get("choice_probabilities"):
                if len(values["choices"]) != len(values["choice_probabilities"]):
                    raise ValueError("Number of choices and probabilities must match")
            if values.get("generator_function"):
                raise ValueError("Cannot specify both choices and generator_function")

        elif not values.get("generator_function"):
            raise ValueError("Must specify either choices or generator_function")

        return values

    def generate(self):
        if self.choices:
            if self.choice_probabilities:
                return random.choices(self.choices, weights=self.choice_probabilities, k=1)[0]
            else:
                return random.choice(self.choices)
        else:
            return self.generator_function()



import pytest
import os
from dotenv import load_dotenv, find_dotenv

from fhir_kindling import FhirServer


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def oidc_server(api_url):
    print(api_url)
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def fhir_server(api_url):
    print(api_url)
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server



import os
import pytest
from dotenv import load_dotenv, find_dotenv

from fhir_kindling import FhirServer

from fhir_kindling.figures import server_summary_plot, plot_resource_field


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def fhir_server(api_url):
    print(api_url)
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


def test_server_summary(fhir_server):
    summary = fhir_server.summary()

    plot_dict = server_summary_plot(summary, show=False)
    assert plot_dict


def test_plot_resource_field(fhir_server):
    patients = fhir_server.query("Patient").all()
    fig = plot_resource_field(patients.resources, field="gender", show=False)
    assert fig
    fig = plot_resource_field(patients.resources, field="gender", plot_type="pie", show=False)
    assert fig



"""Top-level package for fhir-kindling."""

__author__ = """Michael Graf"""
__email__ = 'michael.graf@uni-tuebingen.de'
__version__ = '0.7.2'

from .fhir_server import FhirServer
from .fhir_query import FHIRQuery



from .fhir_query import FHIRQuery
from .query_response import QueryResponse
from .query_parameters import QueryOperators, FHIRQueryParameters, FieldParameter, IncludeParameter, \
    ReverseChainParameter



import plotly.express as px

from fhir_kindling.fhir_server.server_responses import ServerSummary


def server_summary_plot(server_summary: ServerSummary, show: bool = True) -> dict:
    """
    Generate a plotly plot of the server summary displaying the number of available resources on the server.

    Args:
        server_summary: The summary of the server.

    Returns:
        A dict containing the plotly bar plot.
    """

    sorted_resources = sorted(server_summary.available_resources, key=lambda x: x.count, reverse=True)
    fig = px.bar(
        x=[r.resource for r in sorted_resources],
        y=[r.count for r in sorted_resources],
        title="Server Resource Summary"
    )
    if show:
        fig.show()
    return fig.to_dict()




import json

import pandas as pd
import pytest
import os

import xmltodict
from dotenv import load_dotenv, find_dotenv
from fhir.resources.condition import Condition
from fhir.resources.patient import Patient
from pydantic import ValidationError

from fhir_kindling import FhirServer
from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters, IncludeParameter, FieldParameter, \
    ReverseChainParameter, QueryOperators, QueryParameter


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def paginated_xml():
    return """<?xml version="1.0" encoding="UTF-8"?>
<Bundle xmlns="http://hl7.org/fhir">
    <id value="C7GQWQ4ZM4OPHPS4"/>
    <type value="searchset"/>
    <total value="2283"/>
    <link>
        <relation value="self"/>
        <url value="http://localhost:9090/fhir/Patient?_count=50&amp;__t=57&amp;__page-id=C7E64KWQ6FOK7VH7"/>
    </link>
    <link>
        <relation value="next"/>
        <url value="http://localhost:9090/fhir/Patient?_count=50&amp;__t=57&amp;__page-id=C7E64KWQ6FOK7VJL"/>
    </link>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VH7"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VH7"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Carignan"/>
                    <given value="Mirac"/>
                </name>
                <gender value="female"/>
                <birthDate value="1999-11-01"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI2"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI2"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Lougheed"/>
                    <given value="Velmarie"/>
                </name>
                <gender value="male"/>
                <birthDate value="1969-12-01"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI3"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI3"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Solimeno"/>
                    <given value="Cobi"/>
                </name>
                <gender value="male"/>
                <birthDate value="1977-05-19"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI4"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI4"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Lester"/>
                    <given value="Shacoya"/>
                </name>
                <gender value="male"/>
                <birthDate value="1979-03-31"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI5"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI5"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kassin"/>
                    <given value="Ethian"/>
                </name>
                <gender value="male"/>
                <birthDate value="1967-05-17"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI6"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI6"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Weik"/>
                    <given value="Ernel"/>
                </name>
                <gender value="male"/>
                <birthDate value="1975-03-29"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VI7"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VI7"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Einhorn"/>
                    <given value="Kinsly"/>
                </name>
                <gender value="male"/>
                <birthDate value="1989-03-21"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIA"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIA"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kremers"/>
                    <given value="Jessican"/>
                </name>
                <gender value="male"/>
                <birthDate value="1934-06-15"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIB"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIB"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Weininger"/>
                    <given value="Amitava"/>
                </name>
                <gender value="female"/>
                <birthDate value="1925-10-04"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIC"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIC"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Nisly"/>
                    <given value="M'kai"/>
                </name>
                <gender value="female"/>
                <birthDate value="1995-03-01"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VID"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VID"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Mussell"/>
                    <given value="Rogerio"/>
                </name>
                <gender value="female"/>
                <birthDate value="1973-06-17"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIE"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIE"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Muszynski"/>
                    <given value="Natashya"/>
                </name>
                <gender value="male"/>
                <birthDate value="1923-02-08"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIF"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIF"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Soteros"/>
                    <given value="Keyshand"/>
                </name>
                <gender value="male"/>
                <birthDate value="1963-08-07"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIG"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIG"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Wayland"/>
                    <given value="Cellan"/>
                </name>
                <gender value="male"/>
                <birthDate value="1969-02-09"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIH"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIH"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Nemec"/>
                    <given value="Karolann"/>
                </name>
                <gender value="male"/>
                <birthDate value="1996-11-29"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VII"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VII"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Shadik"/>
                    <given value="Lizabet"/>
                </name>
                <gender value="male"/>
                <birthDate value="2003-02-10"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIJ"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIJ"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Farnese"/>
                    <given value="Blimie"/>
                </name>
                <gender value="male"/>
                <birthDate value="1990-01-25"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIK"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIK"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Gwilt"/>
                    <given value="Dur-e-mcknoon"/>
                </name>
                <gender value="male"/>
                <birthDate value="1995-09-15"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIL"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIL"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Bhanu"/>
                    <given value="Beckette"/>
                </name>
                <gender value="female"/>
                <birthDate value="1971-10-05"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIM"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIM"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kleppinger"/>
                    <given value="Austynn"/>
                </name>
                <gender value="female"/>
                <birthDate value="1945-05-05"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIN"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIN"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Umbach"/>
                    <given value="Mirabel"/>
                </name>
                <gender value="female"/>
                <birthDate value="1960-11-16"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIO"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIO"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Vishanu"/>
                    <given value="Mahraj"/>
                </name>
                <gender value="female"/>
                <birthDate value="1985-06-29"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIP"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIP"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Lauth"/>
                    <given value="Conaugh"/>
                </name>
                <gender value="other"/>
                <birthDate value="1937-02-04"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIQ"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIQ"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Samok"/>
                    <given value="Berendina"/>
                </name>
                <gender value="male"/>
                <birthDate value="1988-10-29"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIR"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIR"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kubasik"/>
                    <given value="Glinda"/>
                </name>
                <gender value="male"/>
                <birthDate value="1944-02-06"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIS"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIS"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Blas"/>
                    <given value="Tautiana"/>
                </name>
                <gender value="female"/>
                <birthDate value="1954-08-15"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIT"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIT"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Netto"/>
                    <given value="Emiri"/>
                </name>
                <gender value="female"/>
                <birthDate value="1948-06-21"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIU"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIU"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Dennin"/>
                    <given value="Sassa"/>
                </name>
                <gender value="female"/>
                <birthDate value="1976-04-26"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIV"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIV"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Fietek"/>
                    <given value="Dimitrios"/>
                </name>
                <gender value="female"/>
                <birthDate value="1933-06-05"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIW"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIW"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Pilarski"/>
                    <given value="Keiara"/>
                </name>
                <gender value="female"/>
                <birthDate value="1929-02-21"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIX"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIX"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Gillert"/>
                    <given value="Akalya"/>
                </name>
                <gender value="female"/>
                <birthDate value="1970-08-07"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIY"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIY"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Escalon"/>
                    <given value="Duric"/>
                </name>
                <gender value="female"/>
                <birthDate value="1954-02-08"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VIZ"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VIZ"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Altew"/>
                    <given value="Shelsey"/>
                </name>
                <gender value="male"/>
                <birthDate value="1936-05-24"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ2"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ2"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Schellenberge"/>
                    <given value="Breasia"/>
                </name>
                <gender value="female"/>
                <birthDate value="1954-03-23"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ3"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ3"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Durrwachter"/>
                    <given value="Litta"/>
                </name>
                <gender value="male"/>
                <birthDate value="1936-02-07"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ4"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ4"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Tsukamoto"/>
                    <given value="Jazmen"/>
                </name>
                <gender value="male"/>
                <birthDate value="1964-09-18"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ5"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ5"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kudlacik"/>
                    <given value="Rayonia"/>
                </name>
                <gender value="male"/>
                <birthDate value="1999-05-13"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ6"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ6"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Furrow"/>
                    <given value="Palvit"/>
                </name>
                <gender value="male"/>
                <birthDate value="2002-10-26"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJ7"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJ7"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Kelvin"/>
                    <given value="Darry"/>
                </name>
                <gender value="male"/>
                <birthDate value="1983-03-31"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJA"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJA"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Lebaron"/>
                    <given value="Kelashi"/>
                </name>
                <gender value="male"/>
                <birthDate value="1959-06-01"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJB"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJB"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Heersink"/>
                    <given value="Tamayia"/>
                </name>
                <gender value="female"/>
                <birthDate value="1946-12-22"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJC"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJC"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Quine"/>
                    <given value="Perreault"/>
                </name>
                <gender value="male"/>
                <birthDate value="1982-12-11"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJD"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJD"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Craton"/>
                    <given value="Alonnie"/>
                </name>
                <gender value="female"/>
                <birthDate value="1957-07-18"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJE"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJE"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Dilallo"/>
                    <given value="Linde"/>
                </name>
                <gender value="male"/>
                <birthDate value="1963-05-31"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJF"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJF"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Dolbeare"/>
                    <given value="Furiya"/>
                </name>
                <gender value="male"/>
                <birthDate value="1945-12-12"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJG"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJG"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Pattmon"/>
                    <given value="Crayson"/>
                </name>
                <gender value="male"/>
                <birthDate value="1933-11-28"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJH"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJH"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Carnighan"/>
                    <given value="Ellarose"/>
                </name>
                <gender value="male"/>
                <birthDate value="1956-12-27"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJI"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJI"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Whelihan"/>
                    <given value="Jonette"/>
                </name>
                <gender value="female"/>
                <birthDate value="1930-03-05"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJJ"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJJ"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Petropulos"/>
                    <given value="Zayn-ul-abidin"/>
                </name>
                <gender value="male"/>
                <birthDate value="1930-10-28"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
    <entry>
        <fullUrl value="http://localhost:9090/fhir/Patient/C7E64KWQ6FOK7VJK"/>
        <resource>
            <Patient>
                <id value="C7E64KWQ6FOK7VJK"/>
                <meta>
                    <versionId value="27"/>
                    <lastUpdated value="2021-10-20T18:07:08.130Z"/>
                </meta>
                <name>
                    <family value="Ducci"/>
                    <given value="Kratsas"/>
                </name>
                <gender value="male"/>
                <birthDate value="1987-10-05"/>
            </Patient>
        </resource>
        <search>
            <mode value="match"/>
        </search>
    </entry>
</Bundle>
"""


"""
########################################################################################################################
Test Query Parameters
########################################################################################################################
"""


def test_query_param_abstract():
    query_param = QueryParameter()
    with pytest.raises(NotImplementedError):
        query_param.to_url_param()

    with pytest.raises(NotImplementedError):
        param = QueryParameter.from_url_param("sdada")


def test_field_query_param():
    resource = "Patient"
    field_param = FieldParameter(
        **dict(
            field="active",
            value=True,
            operator=QueryOperators.eq
        )
    )

    assert field_param.to_url_param() == "active=true"

    field_param = FieldParameter(
        **dict(
            field="active",
            value=["hello", "world"],
            operator=QueryOperators.in_
        )
    )

    assert field_param.to_url_param() == "active=hello,world"

    value = 7.3213
    field_param = FieldParameter(
        **dict(
            field="active",
            value=value,
            operator=QueryOperators.gt
        )
    )
    assert field_param.to_url_param() == f"active=gt{value}"

    # list value arg without in operator should fail
    with pytest.raises(ValidationError):
        field_param = FieldParameter(
            **dict(
                field="active",
                value=["hello", "world"],
                operator=QueryOperators.eq
            )
        )

    # in operator only for list value arg
    with pytest.raises(ValidationError):
        field_param = FieldParameter(
            **dict(
                field="active",
                value="hello",
                operator=QueryOperators.in_
            )
        )
    with pytest.raises(ValidationError):
        field_param = FieldParameter(
            **dict(
                field="active",
                value=[32, "sdsa".encode()],
                operator=QueryOperators.in_
            )
        )

    # operator not included in enum
    with pytest.raises(ValidationError):
        field_param = FieldParameter(
            field="active",
            operator="hello",
            value=True
        )

    # parameter generation from url snippet

    param_from_url = FieldParameter.from_url_param("active=true")

    assert param_from_url.field == "active"
    assert param_from_url.operator == QueryOperators.eq
    assert isinstance(param_from_url.value, bool)
    assert param_from_url.value is True

    # list of search params
    param_from_url = FieldParameter.from_url_param("active=hello,world")

    assert param_from_url.field == "active"
    assert param_from_url.operator == QueryOperators.in_
    assert param_from_url.value == ["hello", "world"]

    # integer value
    param_from_url = FieldParameter.from_url_param("integer=gt7")
    assert param_from_url.field == "integer"
    assert param_from_url.operator == QueryOperators.gt
    assert isinstance(param_from_url.value, int)
    assert param_from_url.value == 7

    # float value
    param_from_url = FieldParameter.from_url_param("float=7.3213")
    assert param_from_url.field == "float"
    assert param_from_url.operator == QueryOperators.eq
    assert param_from_url.value == 7.3213

    # list of numbers
    param_from_url = FieldParameter.from_url_param("integers=7,8")
    assert param_from_url.field == "integers"
    assert param_from_url.operator == QueryOperators.in_
    assert param_from_url.value == [7, 8]


def test_include_query_param():
    resource = "Condition"
    search_param = "subject"
    target = "test"

    # conversion to url strings

    # basic include
    include_param = IncludeParameter(
        resource=resource,
        search_param=search_param,
    )
    assert include_param.to_url_param() == f"_include={resource}:{search_param}"

    # with target
    include_param = IncludeParameter(
        resource=resource,
        search_param=search_param,
        target=target
    )
    assert include_param.to_url_param() == f"_include={resource}:{search_param}&{target}"

    # reverse include
    include_param = IncludeParameter(
        resource=resource,
        search_param=search_param,
        target=target,
        reverse=True
    )
    assert include_param.to_url_param() == f"_revinclude={resource}:{search_param}&{target}"

    # include with iterate & target
    include_param = IncludeParameter(
        resource=resource,
        search_param=search_param,
        target=target,
        iterate=True,
    )

    assert include_param.to_url_param() == f"_include:iterate={resource}:{search_param}&{target}"

    # parse from url snippet

    include_param = IncludeParameter.from_url_param(f"_include={resource}:{search_param}")
    assert include_param.resource == resource
    assert include_param.search_param == search_param

    include_param = IncludeParameter.from_url_param(f"_include={resource}:{search_param}&{target}")
    assert include_param.resource == resource
    assert include_param.search_param == search_param
    assert include_param.target == target

    include_param = IncludeParameter.from_url_param(f"_revinclude={resource}:{search_param}&{target}")
    assert include_param.resource == resource
    assert include_param.search_param == search_param
    assert include_param.target == target
    assert include_param.reverse is True

    include_param = IncludeParameter.from_url_param(f"_include:iterate={resource}:{search_param}&{target}")
    assert include_param.resource == resource
    assert include_param.search_param == search_param
    assert include_param.target == target
    assert include_param.iterate is True

    with pytest.raises(ValueError):
        include_param = IncludeParameter.from_url_param(f"_include:iterates={resource}:{search_param}&{target}")


def test_reverse_chain_parameters():
    resource = "Condition"
    reference_param = "patient"
    search_param = "code"
    value = "test"
    query_url = f"_has:{resource}:{reference_param}:{search_param}=test"
    chain_param = ReverseChainParameter(
        resource=resource,
        reference_param=reference_param,
        search_param=search_param,
        operator=QueryOperators.eq,
        value=value
    )

    assert chain_param.to_url_param() == query_url

    param_from_url = ReverseChainParameter.from_url_param(query_url)

    assert param_from_url.resource == chain_param.resource

    query_url = f"_has:{resource}:{reference_param}:{search_param}=netest,test2"
    param_from_url = ReverseChainParameter.from_url_param(query_url)
    assert param_from_url.value == ["test", "test2"]
    assert param_from_url.operator == QueryOperators.not_in


def test_fhir_query_parameters():
    query_params = FHIRQueryParameters(
        resource="Condition"
    )
    assert query_params.to_query_string() == "/Condition?"

    # invalid resource name
    with pytest.raises(ValidationError):
        query_params = FHIRQueryParameters(
            resource="Conditionkdjsaldj"
        )

    url = "/Patient?"
    query_params = FHIRQueryParameters.from_query_string(url)

    assert query_params.resource == "Patient"

    query_url = "/Condition?code=test&_include=Condition:patient&_has:Patient:subject:age=gt:18"
    query_params = FHIRQueryParameters.from_query_string(query_url)

    assert query_params.to_query_string() == query_url


"""
########################################################################################################################
Test query conditions and execution
########################################################################################################################
"""


def test_query_xml(server):
    resp = server.query("Patient", output_format="xml")
    result = resp.limit(100)

    assert result.response
    assert isinstance(result.response, str)

    with pytest.raises(NotImplementedError):
        includes = result.included_resources


def test_query_json(server):
    resp = server.query("Patient", output_format="json")
    result = resp.limit(100)
    assert result.response
    assert isinstance(result.response, dict)
    with pytest.raises(ValueError):
        includes = result.included_resources


def test_query_first(server):
    query = server.query("Patient")
    result = query.first()

    assert len(result.resources) == 1
    assert isinstance(result.resources[0], Patient)


def test_query_limit(server):
    query = server.query("Patient")
    query._count = 30
    result = query.limit(100)
    assert len(result.resources) == 100


def test_query_response_resources(server):
    query = server.query("Patient")
    result = query.all()

    assert len(result.resources) >= 1
    assert isinstance(result.resources[0], Patient)


def test_set_query_string(server, api_url):
    query_resource = "Condition"
    search_param = "subject"

    query = server.query(query_resource)
    query_string = f"/{query_resource}?_include={query_resource}:{search_param}"

    query.set_query_string(query_string)
    query_url = f"{api_url}/{query_resource}?_include={query_resource}:{search_param}&_count=5000&_format=json"
    assert query.query_url == query_url


def test_query_where(server):
    query_resource = "Patient"

    query = server.query(query_resource)

    with pytest.raises(ValueError):
        query.where(field_param="jdsha", filter_dict={"jdsh": "jdsh"})

    with pytest.raises(ValueError):
        query.where()

    with pytest.raises(ValueError):
        query.where(field_param="jdsha", field="dsad")

    with pytest.raises(ValueError):
        query.where(field="jdsha", filter_dict={"jdsh": "jdsh"})

    assert query.resource.resource_type == query_resource

    param = FieldParameter(
        field="name",
        operator=QueryOperators.eq,
        value="test"
    )

    param_dict = dict(
        field="name",
        operator="gt",
        value="test"
    )

    query = query.where(field_param=param)

    assert len(query.query_parameters.resource_parameters) == 1
    assert query.query_parameters.resource_parameters[0].field == "name"

    query = query.where(filter_dict=param_dict)

    assert len(query.query_parameters.resource_parameters) == 2
    assert query.query_parameters.resource_parameters[1].field == "name"
    assert query.query_parameters.resource_parameters[1].operator == QueryOperators.gt

    query = query.where(field="name", operator="lt", value="test")

    assert len(query.query_parameters.resource_parameters) == 3
    assert query.query_parameters.resource_parameters[2].field == "name"
    assert query.query_parameters.resource_parameters[2].operator == QueryOperators.lt

    query = query.where(field="name", operator=QueryOperators.lt, value="test")

    assert len(query.query_parameters.resource_parameters) == 4
    assert query.query_parameters.resource_parameters[2].field == "name"
    assert query.query_parameters.resource_parameters[2].operator == QueryOperators.lt

    with pytest.raises(ValueError):
        query = query.where(field="name", operator="fails", value="test")

    with pytest.raises(ValueError):
        query = query.where(field="name", operator=67321.21, value="test")

    # with pytest.raises(ValueError):
    query = query.where(field="name", operator=QueryOperators.eq, value="test")

    with pytest.raises(ValueError):
        query = query.where(field="name", value="test")

    with pytest.raises(ValidationError):
        param_dict = dict(
            field="name",
            operator="fails",
            value="test"
        )
        query = query.where(filter_dict=param_dict)

    print(query.query_url)

    # execute a valid query
    valid_query_param = FieldParameter(
        field="birthdate",
        operator=QueryOperators.gt,
        value="1930-01-01"
    )
    query = server.query("Patient")
    query = query.where(field_param=valid_query_param)

    response = query.all()

    assert response


def test_query_include(server, api_url):
    query_resource = "Condition"
    search_param = "subject"

    query = server.query(query_resource)

    with pytest.raises(ValueError):
        query = query.include()

    query = query.include(resource=query_resource, reference_param=search_param)

    assert query.query_parameters.include_parameters[0].resource == query_resource

    query_url = f"{api_url}/{query_resource}?_include={query_resource}:{search_param}&_count=5000&_format=json"
    print(query.query_url)
    assert query.query_url == query_url

    include_param = IncludeParameter(
        resource=query_resource,
        search_param=search_param,
        target="criteria",
        reverse=True
    )

    query = query.include(include_param=include_param)
    assert query.query_parameters.include_parameters[1].resource == query_resource
    assert query.query_parameters.include_parameters[1].target == "criteria"

    include_dict = dict(
        resource=query_resource,
        search_param=search_param,
        target="hello",
        reverse=False
    )

    query = query.include(include_dict=include_dict)
    assert query.query_parameters.include_parameters[2].resource == query_resource
    assert query.query_parameters.include_parameters[2].target == "hello"

    with pytest.raises(ValueError):
        query = query.include(include_param=include_param, resource=query_resource)

    with pytest.raises(ValueError):
        query = query.include(include_dict=include_dict, resource=query_resource)

    with pytest.raises(ValueError):
        query = query.include(include_param=include_param, include_dict=include_dict)


def test_query_reverse_chain(server, api_url):
    has_resource = "Condition"
    reference_param = "subject"
    search_param = "code"
    operator = QueryOperators.eq
    value = "test"

    query = server.query("Patient")

    query = query.has(resource=has_resource, reference_param=reference_param, search_param=search_param,
                      operator=operator, value=value)

    assert query

    has_param = ReverseChainParameter(
        resource=has_resource,
        reference_param=reference_param,
        search_param=search_param,
        operator=operator,
        value=7
    )

    has_param_dict = dict(
        resource=has_resource,
        reference_param=reference_param,
        search_param=search_param,
        operator=QueryOperators.in_,
        value=["test1", "test2"]
    )

    query = query.has(has_param=has_param)
    assert query.query_parameters.has_parameters[1].resource == has_resource
    assert query.query_parameters.has_parameters[1].value == 7

    query = query.has(has_param_dict=has_param_dict)

    assert query.query_parameters.has_parameters[2].resource == has_resource
    assert query.query_parameters.has_parameters[2].operator == QueryOperators.in_
    assert query.query_parameters.has_parameters[2].value == ["test1", "test2"]

    with pytest.raises(ValueError):
        query = query.has(has_param=has_param, has_param_dict=has_param_dict)

    with pytest.raises(ValueError):
        query = query.has(has_param=has_param, resource="test")

    with pytest.raises(ValueError):
        query = query.has(resource="has_param", has_param_dict=has_param_dict)

    with pytest.raises(ValueError):
        query = query.has()


"""
########################################################################################################################
Test Query Response
########################################################################################################################
"""


def test_query_response(server):
    query_resource = "Patient"

    query = server.query(query_resource)
    query._count = 50

    response = query.all()

    assert len(response.resources) >= 2
    assert response.resources[0].resource_type == query_resource


def test_query_response_include(server):
    query_resource = "Condition"
    search_param = "subject"

    query = server.query(query_resource)

    with pytest.raises(ValueError):
        query = query.include()

    query = query.include(resource=query_resource, reference_param=search_param)
    response = query.all()

    assert response.included_resources
    assert response.resources
    assert response.resources[0].resource_type == query_resource


def test_query_response_save(server):
    xml_file = "test_query_response_save.xml"
    json_file = "test_query_response_save.json"
    csv_file = "test_query_response_save.csv"

    # test saving single resource and xml
    query_resource = "Condition"
    query = server.query(query_resource, output_format="xml")
    response = query.limit(100)
    response.save(xml_file, format="xml")

    with pytest.raises(NotImplementedError):
        response.save(xml_file, format="json")

    assert os.path.isfile(xml_file)
    assert xmltodict.parse(open(xml_file, "r").read()) == xmltodict.parse(response.response)

    os.remove(xml_file)

    # invalid format
    with pytest.raises(ValueError):
        response.save(xml_file, format="invalid")

    # test with included resources and json
    query_resource = "Condition"
    search_param = "subject"

    query = server.query(query_resource)

    with pytest.raises(ValueError):
        query = query.include()

    query = query.include(resource=query_resource, reference_param=search_param)
    response = query.all()

    response.save(json_file, format="json")

    assert os.path.isfile(json_file)
    assert json.loads(open(json_file).read())

    os.remove(json_file)

    # save to csv
    patient_response = server.query("Patient").all()
    patient_response.save(csv_file, format="csv")

    assert os.path.isfile(csv_file)
    assert pd.read_csv(csv_file).shape[0] > 0

    # os.remove(csv_file)

    # save to csv with included resource
    response.save(csv_file, format="csv")

    assert os.path.isfile(csv_file)
    assert pd.read_csv(csv_file).shape[0] > 0

    included_patients_csv = "test_query_response_save_included_Patient.csv"
    assert os.path.isfile(included_patients_csv)
    assert pd.read_csv(included_patients_csv).shape[0] > 0

    os.remove(csv_file)
    os.remove(included_patients_csv)


def test_query_response_to_dfs(server):
    # test with included resources and json
    query_resource = "Condition"
    search_param = "subject"

    query = server.query(query_resource)

    with pytest.raises(ValueError):
        query = query.include()

    query = query.include(resource=query_resource, reference_param=search_param)
    response = query.all()

    print(query)

    with pytest.raises(ValueError):
        response.to_dfs(format="invalid")

    dfs = response.to_dfs()

    assert len(dfs) > 1
    assert dfs[0].shape[0] > 0

    print(response)



import json
import os
from typing import List, Union, Tuple
import requests
from fhir.resources import FHIRAbstractModel
from requests import Response
import requests.auth
from fhir.resources.resource import Resource
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.capabilitystatement import CapabilityStatement
from fhir.resources.reference import Reference
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from requests_oauthlib import OAuth2Session
import fhir.resources
import re
from oauthlib.oauth2 import BackendApplicationClient
import pendulum

from fhir_kindling.fhir_query import FHIRQuery
from fhir_kindling.fhir_query.query_response import QueryResponse
from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters
from fhir_kindling.fhir_server.auth import generate_auth
from fhir_kindling.fhir_server.server_responses import ResourceCreateResponse, BundleCreateResponse, ServerSummary


class FhirServer:

    def __init__(self, api_address: str, username: str = None, password: str = None, token: str = None,
                 client_id: str = None, client_secret: str = None, oidc_provider_url: str = None,
                 auth: requests.auth.AuthBase = None, headers: dict = None,
                 fhir_server_type: str = "hapi"):
        """
        Initialize a FHIR server connection
        Args:
            api_address: the base endpoint of the fhir server api
            username: username for basic auth
            password: password for basic auth
            token: token for static token auth
            client_id: client id for oauth2
            client_secret: client secret for oauth2
            oidc_provider_url: provider url for oauth2
            auth: optional auth object to authenticate against a server
            headers: optional additional headers to be added to the session
            fhir_server_type: type of fhir server (hapi, blaze, etc.)
        """

        # server definition values
        self.fhir_server_type = fhir_server_type
        self.api_address = self.validate_api_address(api_address)
        self._meta_data = None
        self.auth = None

        # possible basic auth class vars
        self.username = username
        self.password = password

        # token oidc auth class vars
        self.token = token
        self.client_id = client_id
        self.client_secret = client_secret
        self.oidc_provider_url = oidc_provider_url
        self.token_expiration = None

        self._auth = auth
        self._headers = headers

        # setup the session
        self.session = requests.Session()
        self._setup()

    def query(self, resource: Union[Resource, FHIRAbstractModel, str] = None,
              query_parameters: FHIRQueryParameters = None,
              output_format: str = "json", count: int = 5000) -> FHIRQuery:
        """
        Initialize a FHIR query against the server with the given resource or query parameters

        Args:
            output_format: the output format to request from the fhir server (json or xml) defaults to json
            query_parameters: optionally pass in a query parameters object to use for the query
            resource: the FHIR resource to query from the server
            count: the number of resources requested per page

        Returns: a FHIRQuery object that can be further modified with filters and conditions before being executed
        against the server
        """
        if resource:
            return FHIRQuery(self.api_address, resource, auth=self.auth, session=self.session,
                             output_format=output_format, count=count)
        else:
            return FHIRQuery(self.api_address, auth=self.auth, session=self.session, query_parameters=query_parameters,
                             count=count, output_format=output_format)

    def raw_query(self, query_string: str, output_format: str = "json") -> FHIRQuery:
        """
        Execute a raw query string against the server

        Args:
            query_string:
            output_format:
            limit:
            count:

        Returns:

        """

        query_parameters = FHIRQueryParameters.from_query_string(query_string)
        query = FHIRQuery(
            self.api_address,
            resource=query_parameters.resource,
            query_parameters=query_parameters,
            session=self.session,
            output_format=output_format
        )
        return query

    def add(self, resource: Union[Resource, dict]) -> ResourceCreateResponse:
        """
        Upload a resource to the server

        Args:
            resource: dictionary containing the resource or FHIR resource object to be uploaded to the server

        Returns:

        """
        # parse the resource given as dictionary into a fhir resource model
        if isinstance(resource, dict):
            resource_type = resource.get("resourceType")
            if not resource_type:
                raise ValueError("No resource type defined in resource dictionary")
            resource = fhir.resources.construct_fhir_element(resource_type, resource)
        else:
            resource = resource.validate(resource)
        response = self._upload_resource(resource)
        response.raise_for_status()

        return ResourceCreateResponse(server_response_dict=dict(response.headers), resource=resource)

    def add_all(self, resources: List[Union[Resource, dict]]) -> BundleCreateResponse:
        """
        Upload a list of resources to the server, after packaging them into a bundle
        Args:
            resources: list of resources to upload to the server, either dictionary or FHIR resource objects

        Returns: Bundle create response from the fhir server

        """
        bundle = self._make_bundle_from_resource_list(resources)
        response = self._upload_bundle(bundle)
        return response

    def add_bundle(self, bundle: Union[Bundle, dict, str], validate_entries: bool = True) -> BundleCreateResponse:
        # create bundle and validate it
        if isinstance(bundle, dict):
            bundle = Bundle(**bundle)
        elif isinstance(bundle, str):
            bundle = Bundle.parse_raw(bundle)
        else:
            bundle = Bundle.validate(bundle)
        # check that all entries are bundle requests with methods post/put
        if validate_entries:
            self._validate_upload_bundle_entries(bundle.entry)

        transaction_response = self._upload_bundle(bundle)
        return transaction_response

    def update(self, resources: List[Union[FHIRResourceModel, dict]]):
        """
        Update a list of resources that exist on the server
        Args:
            resources: List of updated resources coming to send to the server

        Returns: Bundle update response from the fhir server

        """

        # todo check that the resources exist and have an id?
        # batch update transaction
        update_bundle = Bundle.construct()
        update_bundle.type = "transaction"
        update_bundle.entry = []
        for resource in resources:
            if isinstance(resource, dict):
                resource_type = resource.get("resourceType")
                if not resource_type:
                    raise ValueError("No resource type defined in resource dictionary")
                resource = fhir.resources.construct_fhir_element(resource_type, resource)
            elif isinstance(resource, FHIRResourceModel):
                resource = resource.validate(resource)
            else:
                raise ValueError(f"Invalid resource type {type(resource)}")
            # make the transaction entry
            entry = self._make_transaction_entry(resource.relative_path(), "PUT")
            # add the resource to the entry
            entry.resource = resource
            # validate the entry and append it to the bundle
            update_bundle.entry.append(BundleEntry(**entry.dict()))

        # validate bundle
        update_bundle = Bundle(**update_bundle.dict())
        response = self.session.post(self.api_address, data=update_bundle.json())
        response.raise_for_status()

        return response.json()

    def delete(self,
               resources: List[Union[FHIRResourceModel, dict]] = None,
               references: List[Union[str, Reference]] = None,
               query: FHIRQuery = None):
        """
        Delete resources from the server. Either resources or references must be specified.
        Args:
            resources: Resources coming from the server containing an id to delete
            references: references {Resource}/{id} to delete
            query: query to use to find resources to delete

        Returns: Bundle delete response from the fhir server

        """
        if resources and references:
            raise ValueError("Cannot delete based on resources and references at the same time")
        if resources and query:
            raise ValueError("Cannot delete based on resources and query at the same time")
        if references and query:
            raise ValueError("Cannot delete based on references and query at the same time")

        transaction_bundle = self._make_delete_transaction(resources, references, query)

        response = self.session.post(self.api_address, json=transaction_bundle.dict())
        response.raise_for_status()

        return response

    def transfer(self, target_server: 'FhirServer', query: FHIRQuery = None,
                 bundle: Union[Bundle, dict, str, bytes] = None):
        if query and bundle:
            raise ValueError("Cannot transfer based on query and bundle at the same time")
        if query:
            # todo execute query, get bundle, parse references, upload bundle
            raise NotImplementedError("Transfer by query is not implemented yet")

        elif bundle:
            if isinstance(bundle, dict):
                bundle = Bundle(**bundle)
            elif isinstance(bundle, str):
                bundle = Bundle.parse_raw(bundle)

        response = self._transfer_bundle(bundle, target_server)

    def _transfer_bundle(self, bundle: Bundle, target_server: 'FhirServer') -> Bundle:
        # todo
        staged_upload = self._resolve_references(bundle)

    def _resolve_references(self, bundle: Bundle):
        # todo
        pass

    @classmethod
    def from_env(cls, no_auth: bool = False):
        api_address = _api_address_from_env()
        server_type = os.getenv("FHIR_SERVER_TYPE")

        if no_auth:
            return FhirServer(api_address=api_address, fhir_server_type=server_type)
        else:
            env_auth = _auth_info_from_env()
            # static token
            if isinstance(env_auth, str):
                return cls(api_address=api_address, token=env_auth, fhir_server_type=server_type)
            # username and password
            elif isinstance(env_auth, tuple) and len(env_auth) == 2:
                return cls(api_address=api_address, username=env_auth[0], password=env_auth[1],
                           fhir_server_type=server_type)
            # oauth2/oidc
            elif isinstance(env_auth, tuple) and len(env_auth) == 3:
                return cls(api_address=api_address, client_id=env_auth[0], client_secret=env_auth[1],
                           oidc_provider_url=env_auth[2], fhir_server_type=server_type)
            else:
                raise EnvironmentError("Authentication information could not be loaded from environment")

    def _make_delete_transaction(self, resources: List[Union[Resource, dict]] = None,
                                 references: List[Union[str, Reference]] = None, query: QueryResponse = None) -> Bundle:
        delete_bundle = Bundle.construct()
        delete_bundle.type = "transaction"
        delete_bundle.entry = []

        if resources:
            print(resources)
            if isinstance(resources[0], dict):
                resources = [fhir.resources.construct_fhir_element(
                    resource.get("resourceType"),
                    resource
                )
                    for resource in resources
                ]

            delete_references = [res.relative_path() for res in resources]
        elif references:
            if isinstance(references[0], Reference):
                references = [ref.reference for ref in references]
            delete_references = references
        elif query:

            delete_references = [resource.relative_path() for resource in query.resources]
        else:
            raise ValueError("No resources or references provided")

        bundle_entries = [self._make_transaction_entry(ref, method="DELETE") for ref in delete_references]
        delete_bundle.entry = bundle_entries
        return delete_bundle

    @staticmethod
    def _make_transaction_entry(reference: str, method: str = "POST") -> BundleEntry:
        entry = BundleEntry.construct()
        entry.request = BundleEntryRequest(
            **{
                "method": method,
                "url": reference
            }
        )

        return entry

    def _make_bundle_from_resource_list(self, resources: List[Union[FHIRAbstractModel, dict]]) -> Bundle:
        upload_bundle = Bundle.construct()
        upload_bundle.type = "transaction"
        upload_bundle.entry = []

        # initialize fhir model instances from dict
        if isinstance(resources[0], dict):
            resources = [fhir.resources.construct_fhir_element(
                resource.get("resourceType"),
                resource
            )
                for resource in resources
            ]
        for resource in resources:
            entry = self._make_bundle_request_entry(resource)
            upload_bundle.entry.append(entry)

        return upload_bundle

    @staticmethod
    def _validate_upload_bundle_entries(entries: List[BundleEntry]):
        for i, entry in enumerate(entries):
            if entry.request.method.lower() not in ["post", "put"]:
                raise ValueError(f"Entry {i}:  method is not in [post, put]")

    @staticmethod
    def _make_bundle_request_entry(resource: FHIRAbstractModel) -> BundleEntry:
        entry = BundleEntry().construct()
        entry.request = BundleEntryRequest(
            **{
                "method": "POST",
                "url": resource.get_resource_type()
            }
        )
        entry.resource = resource

        return entry

    def _upload_bundle(self, bundle: Bundle) -> BundleCreateResponse:
        r = self.session.post(url=self.api_address, data=bundle.json(return_bytes=True))
        r.raise_for_status()
        bundle_response = BundleCreateResponse(r, bundle)
        return bundle_response

    def _upload_resource(self, resource: Resource) -> Response:
        url = self.api_address + "/" + resource.get_resource_type()
        r = requests.post(url=url, headers=self.headers, auth=self.auth, json=resource.dict())
        return r

    def _get_meta_data(self):
        url = self.api_address + "/metadata"
        r = self.session.get(url)
        r.raise_for_status()
        response = r.json()
        self._meta_data = response

    def _setup(self):
        self.auth = self._validate_auth()
        self.session.auth = self.auth
        self.session.headers.update(self.headers)

    @property
    def capabilities(self) -> CapabilityStatement:
        if not self._meta_data:
            self._get_meta_data()
        return CapabilityStatement(**self._meta_data)

    @property
    def rest_resources(self) -> List[str]:
        return [capa.type for capa in self.capabilities.rest[0].resource]

    def summary(self) -> ServerSummary:
        summary = self._make_server_summary()
        return summary

    def _make_server_summary(self) -> ServerSummary:
        resources = []
        summary = {
            "name": self.api_address
        }
        for resource in self.rest_resources:
            url = self.api_address + "/" + resource + "?_summary=count"
            r = self.session.get(url)
            r.raise_for_status()

            resource_dict = {
                "resource": resource,
                "count": r.json().get("total")
            }
            resources.append(resource_dict)

        summary["resources"] = resources
        summary = ServerSummary(**summary)
        return summary

    @property
    def headers(self):
        headers = {"Content-Type": "application/fhir+json"}
        if self._headers:
            headers.update(self._headers)
        return headers

    def _get_oidc_token(self):
        # get a new token if it is expired or not yet set
        if (self.token_expiration and pendulum.now() > self.token_expiration) or not self.token:
            print("Requesting new token")
            client = BackendApplicationClient(client_id=self.client_id)
            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(
                token_url=self.oidc_provider_url,
                client_secret=self.client_secret,
                client_id=self.client_id
            )
            self.token = token["access_token"]
            self.token_expiration = pendulum.now() + pendulum.duration(seconds=token["expires_in"])

    @staticmethod
    def validate_api_address(api_address: str) -> str:
        """
        Validate that api address is well formed and remove trailing / if present
        https://stackoverflow.com/a/7160778/3838313

        Args:
            api_address: base endpoint for the fhir rest api of the server

        Returns:
            the validated api address

        """

        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'[A-Za-z0-9_-]*|'  # single word with hyphen/underscore for docker
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        if re.match(regex, api_address):
            if api_address[-1] == "/":
                api_address = api_address[:-1]
            return api_address

        else:
            raise ValueError(f"Malformed API URL: {api_address}")

    def _validate_auth(self) -> Union[requests.auth.AuthBase, None]:
        if self._auth and (self.username or self.password):
            raise ValueError(
                "Only one authentication method can be used, auth object and username/password are mutually exclusive")
        if self._auth and self.token:
            raise ValueError("Only one authentication method can be used, auth object and token are mutually exclusive")
        if self._auth and (self.client_id or self.client_secret or self.oidc_provider_url):
            raise ValueError(
                "Only one authentication method can be used, auth object and OIDC auth are mutually exclusive")

        if self.token and (self.client_id or self.client_secret or self.oidc_provider_url):
            raise ValueError("Only one authentication method can be used, token and OIDC auth are mutually exclusive")

        if self.token and (self.username or self.password):
            raise ValueError(
                "Only one authentication method can be used, token and username/password are mutually exclusive")

        if (self.username or self.password) and (self.client_id or self.client_secret or self.oidc_provider_url):
            raise ValueError("Only one authentication method can be used, username/password and "
                             "OIDC auth are mutually exclusive")

        if self._auth:
            return self._auth
        # OIDC authentication
        if self.client_id:
            self._get_oidc_token()
            return generate_auth(token=self.token)
        # basic or static token authentication
        elif self.username and self.password:
            return generate_auth(self.username, self.password)

        # static token auth
        elif self.token:
            return generate_auth(token=self.token)
        else:
            return None

    def __repr__(self):

        return f"<FHIRServer api_address={self.api_address}, auth={self.auth}, headers={self.headers}>"


def _api_address_from_env() -> str:
    # load FHIR_API_URL
    api_url = os.getenv("FHIR_API_URL")
    if not api_url:
        api_url = os.getenv("FHIR_SERVER_URL")
    if not api_url:
        raise EnvironmentError("No FHIR api address specified")
    return FhirServer.validate_api_address(api_url)


def _auth_info_from_env() -> Union[str, Tuple[str, str], Tuple[str, str, str]]:
    # First try to load basic auth information
    username = os.getenv("FHIR_USER")
    # Static token auth
    token = os.getenv("FHIR_TOKEN")
    # oauth2/oidc authentication
    client_id = os.getenv("CLIENT_ID")
    client_secret = os.getenv("CLIENT_SECRET")
    oidc_provider_url = os.getenv("OIDC_PROVIDER_URL")

    if username and token:
        raise EnvironmentError("Conflicting auth information, bother username and token present.")
    if username and client_id:
        raise EnvironmentError("Conflicting auth information, bother username and client id present")
    if token and client_id:
        raise EnvironmentError("Conflicting auth information, bother static token and client id present")

    if username:
        password = os.getenv("FHIR_PW")
        if not password:
            raise EnvironmentError(f"No password specified for user: {username}")
        else:
            print(f"Basic auth environment info found -> ({username}:******)")
            return username, password
    if token:
        print("Found static auth token")
        return token

    if client_id and not client_secret:
        raise EnvironmentError("Insufficient auth information, client id specified but no client secret found.")

    if (client_id and client_secret) and not oidc_provider_url:
        raise EnvironmentError("Insufficient auth information, client id and secret "
                               "specified but no provider URL found")
    if client_id and client_secret and oidc_provider_url:
        print(f"Found OIDC auth configuration for client <{client_id}> with provider {oidc_provider_url}")
        return client_id, client_secret, oidc_provider_url





from typing import List
import plotly.graph_objects as go
from fhir.resources.resource import Resource
import pandas as pd


def plot_resource_field(resources: List[Resource], field: str, title: str = None, plot_type: str = "bar",
                        show: bool = True) -> go.Figure:
    """
    Plot a field of a resource.

    :param resources: Resources for which to plot the field
    :param field: Field to plot
    :param title: Title of the plot
    :param plot_type: Type of plot to use. Options are: bar, histogram, pie
    :param show: Show the plot
    :return: Plotly figure
    """
    if title is None:
        title = f"{field} for {resources[0].resource_type}"

    figure = go.Figure()
    figure.update_layout(title_text=title, title_x=0.5)
    values = [resource.dict().get(field) for resource in resources]
    # convert to series and get value counts
    val_counts = pd.Series(values).value_counts()
    if plot_type == "bar":
        figure.add_trace(go.Bar(
            x=val_counts.index,
            y=val_counts.values,
            name=field
        ))
    elif plot_type == "pie":
        figure.add_trace(go.Pie(
            labels=val_counts.index,
            values=val_counts.values,
            name=field
        ))
    if show:
        figure.show()

    return figure



import os

import pytest
from dotenv import load_dotenv, find_dotenv

from fhir_kindling import FhirServer
from fhir_kindling.serde.flatten import flatten_resource, flatten_resources


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


def test_flatten_resource(server):
    patient = server.query("Patient").limit(1).resources[0]
    flat_patient = flatten_resource(patient)

    condition = server.query("Condition").limit(1).resources[0]
    flat_condition = flatten_resource(condition)

    # observation = server.query("Observation").limit(1).resources[0]
    # flat_obs = flatten_resource(observation)


def test_resources_to_csv(server):
    patients = server.query("Patient").all().resources
    patients_df = flatten_resources(patients)
    patients_df.to_csv("patients.csv")

    if os.path.exists("patients.csv"):
        os.remove("patients.csv")

    conditions = server.query("Condition").all().resources
    conditions_df = flatten_resources(conditions)
    conditions_df.to_csv("conditions.csv")

    if os.path.exists("conditions.csv"):
        os.remove("conditions.csv")



import pytest
from fhir_kindling.fhir_server.auth import generate_auth, load_environment_auth_vars, BearerAuth
from requests.auth import HTTPBasicAuth
from unittest import mock
import os


def test_generate_auth():
    basic_auth = generate_auth(username="basic", password="auth")

    assert isinstance(basic_auth, HTTPBasicAuth)

    bearer_auth = generate_auth(token="test_token")

    assert isinstance(bearer_auth, BearerAuth)

    # Conflicting auth information should raise an error
    with pytest.raises(ValueError):
        generate_auth(username="test", password="conflicting", token="auth")

    # No auth information should give an error
    with pytest.raises(ValueError):
        generate_auth()
    with pytest.raises(ValueError):
        generate_auth(username="test")


@mock.patch.dict(os.environ, {"FHIR_USER": "test", "FHIR_PW": "password", "FHIR_TOKEN": "fhir-token"})
def test_load_env_var_auth():
    username, password, token = load_environment_auth_vars()

    assert username == "test"
    assert password == "password"
    assert token == "fhir-token"


@mock.patch.dict(os.environ, {"FHIR_USER": "test", "FHIR_PW": "password", "FHIR_TOKEN": "fhir-token"})
def test_env_var_conflicts():
    with pytest.raises(ValueError):
        generate_auth(load_env=True)


@mock.patch.dict(os.environ, {"FHIR_USER": "test", "FHIR_PW": "password"})
def test_env_var_basic():
    basic_auth = generate_auth(load_env=True)
    assert isinstance(basic_auth, HTTPBasicAuth)



from typing import Union, List
from collections import MutableMapping

import pandas as pd
from fhir.resources import FHIRAbstractModel
from fhir.resources.fhirresourcemodel import FHIRResourceModel


def flatten_resources(resources: Union[List[FHIRResourceModel], List[FHIRAbstractModel]]) -> pd.DataFrame:
    """
    Flatten a list of resources of a single resource type into a dataframe.
    Args:
        resources: list of resources

    Returns: pandas dataframe with columns corresponding to the flatten ed resource keys

    """

    flat_resources = []
    for resource in resources:
        flat_resources.append(flatten_resource(resource))
    return pd.DataFrame.from_records(flat_resources)


def flatten_resource(resource: Union[FHIRResourceModel, FHIRAbstractModel]) -> dict:
    """
    Flatten a resource into a single level dictionary.
    Args:
        resource: Fhir resource to convert to single level dict

    Returns: dictionary with only single level keys containing the fields of the resource

    """
    resource_dict = resource.dict(exclude_none=True)
    flat_dict = flatten_dict(resource_dict)
    return flat_dict


def flatten_dict(d, parent_key='', sep='_') -> dict:
    """
    Flatten a nested dictionary into a single level dictionary.

    Extension of https://stackoverflow.com/a/6027615/3838313 capable of handling lists
    Args:
        d: dictionary to flatten
        parent_key: parent key user for recursion
        sep: separator for the nested keys

    Returns:

    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        elif isinstance(v, list):
            for i, item in enumerate(v):
                if isinstance(item, MutableMapping):
                    items.extend(flatten_dict(item, f"{new_key}_{i}", sep=sep).items())
                else:
                    items.append((f"{new_key}_{i}", item))
        else:
            items.append((new_key, v))
    return dict(items)



from typing import Union
from fhir.resources.resource import Resource
from fhir.resources.bundle import Bundle
from fhir.resources.fhirresourcemodel import FHIRResourceModel
import fhir.resources
import requests
import requests.auth

from fhir_kindling.fhir_query.query_response import QueryResponse
from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters, IncludeParameter, FieldParameter, \
    ReverseChainParameter, QueryOperators


class FHIRQuery:
    def __init__(self,
                 base_url: str,
                 resource: Union[FHIRResourceModel, fhir.resources.FHIRAbstractModel, str] = None,
                 query_parameters: FHIRQueryParameters = None,
                 auth: requests.auth.AuthBase = None,
                 session: requests.Session = None,
                 output_format: str = "json",
                 count: int = 5000):

        self.base_url = base_url

        # Set up the requests session with auth and headers
        self.auth = auth
        if session:
            self.session = session
        else:
            self._setup_session()

        # initialize the resource
        if isinstance(resource, str):
            self.resource = fhir.resources.get_fhir_model_class(resource)
        else:
            self.resource = resource

        self.resource = self.resource.construct()
        self.output_format = output_format
        self._includes = None
        self._limit = None
        self._count = count
        self._query_response: Union[Bundle, str] = None
        if query_parameters:
            self.query_parameters = query_parameters
        else:
            self.query_parameters: FHIRQueryParameters = FHIRQueryParameters(resource=self.resource.resource_type)

    def where(self,
              field: str = None,
              operator: Union[QueryOperators, str] = None,
              value: Union[int, float, bool, str, list] = None,
              field_param: FieldParameter = None,
              filter_dict: dict = None
              ) -> 'FHIRQuery':
        """
        Add search conditions regarding a specific field of the queried resource.
        Conditions can be added via FieldParameter class instance, via a dictionary or specifying condition via this
        method's parameter arguments (field, operator, value).
        Args:
            field_param: Instance of FieldParameter defining the field to filter for.
            filter_dict: dictionary containing the field to search for and the value to filter for.
            field: string specifier of the field to fileter for
            operator: comparison operator either as string or QueryOperators
            value: comparison value.

        Returns:

        """

        # evaluate arguments
        if field_param and filter_dict:
            raise ValueError("Cannot use both field_param and filter_dict")
        elif field_param and (field or operator or value):
            raise ValueError("Cannot use both field_param and kv parameters")
        elif filter_dict and (field or operator or value):
            raise ValueError("Cannot use both filter_dict and kv parameters")

        # create field parameters from the different argument options
        if isinstance(field_param, FieldParameter):
            added_query_param = field_param

        elif isinstance(filter_dict, dict):
            # todo allow for multiple filter_dicts/multiple parameters in dict
            added_query_param = FieldParameter(**filter_dict)
        elif field:
            if not (operator or operator == "") and value:
                kv_error_message = f"\n\tField: {field}\n\tOperator: {operator}\n\tValue: {value}"
                raise ValueError(f"Must provide operator and search value when using kv parameters.{kv_error_message}")
            else:

                if isinstance(operator, str):
                    operator = QueryOperators(operator)
                if isinstance(operator, QueryOperators):
                    print(f"Operator: {operator}")
                    operator = operator
                else:
                    raise ValueError(f"Operator must be a string or QueryOperators. Got {operator}")
                added_query_param = FieldParameter(field=field, operator=operator, value=value)

        else:
            raise ValueError("Must provide a valid instance of either field_param or filter_dict or the kv parameters")

        query_field_params = self.query_parameters.resource_parameters
        if isinstance(query_field_params, list) and len(query_field_params) > 0:
            query_field_params.append(added_query_param)
        else:
            query_field_params = [added_query_param]

        self.query_parameters.resource_parameters = query_field_params

        return self

    def include(self,
                resource: str = None,
                reference_param: str = None,
                target: str = None,
                reverse: bool = False,
                include_dict: dict = None,
                include_param: IncludeParameter = None
                ) -> 'FHIRQuery':

        """
        Specify resources related to the queried resource, which should be included in the query results.

        Args:
            resource: name of the resource from which to include related resources, has to match the main resource
                of the query
            reference_param: the reference parameter to search for
            target: further specification of the reference parameter to search for
            reverse: whether to consider reverse references
            include_dict: dictionary container the include parameters
            include_param: instance of IncludeParameter defining the include parameters

        Returns:
            Updated query instance with an added include parameter

        """

        if include_dict and include_param:
            raise ValueError("Cannot use both include_dict and include_param")
        elif include_dict and (resource or reference_param or target):
            raise ValueError("Cannot use both include_dict and kv parameters")
        elif include_param and (resource or reference_param or target):
            raise ValueError("Cannot use both include_param and kv parameters")

        if isinstance(include_dict, dict):
            added_include_param = IncludeParameter(**include_dict)
        elif isinstance(include_param, IncludeParameter):
            added_include_param = include_param
        elif resource and reference_param:
            added_include_param = IncludeParameter(resource=resource, search_param=reference_param, target=target,
                                                   reverse=reverse)
        else:
            raise ValueError(
                "Must provide a valid instance of either include_dict or include_param or the kv parameters")

        query_include_params = self.query_parameters.include_parameters
        if isinstance(query_include_params, list) and len(query_include_params) > 0:
            query_include_params.append(added_include_param)
        else:
            query_include_params = [added_include_param]

        self.query_parameters.include_parameters = query_include_params

        return self

    def has(self,
            resource: str = None,
            reference_param: str = None,
            search_param: str = None,
            operator: QueryOperators = None,
            value: Union[int, float, bool, str, list] = None,
            has_param_dict: dict = None,
            has_param: ReverseChainParameter = None
            ) -> 'FHIRQuery':
        """
        Specify query parameters for other resources that are referenced by the queried, only the resources whose
        referenced resources match the specified search criteria are included in the results.

        Args:
            resource: type of resource that references the selected resource
            reference_param: name of the field of the related resource that defines the relation
            search_param: field of the resource to compare with the given value using the given query operator
            operator: comparison operator, one of QueryOperators
            value: the value to compare the field to
            has_param_dict: dictionary containing the required reverse chain parameters as keys
            has_param: instance of ReverseChainParameter object

        Returns:
            Updated query object with an added ReverseChainParameter

        """

        # validate method input
        if has_param_dict and has_param:
            raise ValueError("Cannot use both has_param_dict and has_param")
        elif has_param_dict and (resource or reference_param or search_param or operator or value):
            raise ValueError("Cannot use both has_param_dict and kv parameters")
        elif has_param and (resource or reference_param or search_param or operator or value):
            raise ValueError("Cannot use both has_param and kv parameters")

        # parse ReverseChainParameter from method input
        if isinstance(has_param_dict, dict):
            added_has_param = ReverseChainParameter(**has_param_dict)
        elif isinstance(has_param, ReverseChainParameter):
            added_has_param = has_param
        elif resource and reference_param and search_param and operator and value:
            added_has_param = ReverseChainParameter(resource=resource, reference_param=reference_param,
                                                    search_param=search_param, operator=operator, value=value)
        else:
            raise ValueError(
                "Either has_param_dict, a parameter instance or a valid set of kv parameters must be provided")

        query_has_params = self.query_parameters.has_parameters
        if isinstance(query_has_params, list) and len(query_has_params) > 0:
            query_has_params.append(added_has_param)
        else:
            query_has_params = [added_has_param]
        self.query_parameters.has_parameters = query_has_params

        return self

    def all(self):
        """
        Execute the query and return all results matching the query parameters.
        Returns:
            QueryResponse object containing all resources matching the query, as well os optional included
            resources.

        """
        self._limit = None
        return self._execute_query()

    def limit(self, n: int):
        """
        Execute the query and return the first n results matching the query parameters.
        Args:
            n: number of resources to return

        Returns:
            QueryResponse object containing the first n resources matching the query, as well os optional included
            resources.

        """
        self._limit = n
        return self._execute_query()

    def first(self):
        """
        Return the first resource matching the query parameters.
        Returns:
            QueryResponse object containing the first resource matching the query

        """
        self._limit = 1
        return self._execute_query()

    def set_query_string(self, raw_query_string: str):
        """
        Use a raw query string to set the query parameters.
        e.g. /Patient?_id=123&_lastUpdated=gt2019-01-01

        Args:
            raw_query_string: Query string to set the query parameters

        Returns:
            Query object with the query parameters set based on the raw query string

        """
        query_parameters = FHIRQueryParameters.from_query_string(raw_query_string)
        self.query_parameters = query_parameters
        return self

    @property
    def query_url(self) -> str:
        """
        Display the query URL that will be used to execute the query.

        Returns:

        """
        return self._make_query_string()

    def _setup_session(self):
        self.session = requests.Session()
        self.session.auth = self.auth
        self.session.headers.update({"Content-Type": "application/fhir+json"})

    def _execute_query(self):
        r = self.session.get(self.query_url)
        r.raise_for_status()
        response = QueryResponse(
            session=self.session,
            response=r,
            query_params=self.query_parameters,
            output_format=self.output_format,
            limit=self._limit
        )
        return response

    def _make_query_string(self):
        query_string = f"{self.base_url}{self.query_parameters.to_query_string()}"
        if self._limit:
            query_string += f"&_count={self._limit}"
        else:
            query_string += f"&_count={self._count}"
        query_string += f"&_format={self.output_format}"

        return query_string

    def __repr__(self):
        return f"<FHIRQuery(resource={self.resource}, url={self.query_url}>"



from .resource_generator import ResourceGenerator, FieldValue, GeneratorParameters
from .patient import PatientGenerator
from .dataset import DatasetGenerator
from .field_generator import FieldGenerator



from uuid import uuid4
from typing import Tuple, List
from pathlib import Path

import pendulum
import pandas as pd
import random
from tqdm import tqdm
from fhir.resources.reference import Reference
from fhir.resources.patient import Patient
from fhir.resources.humanname import HumanName


class PatientGenerator:
    def __init__(self,
                 n: int,
                 age_range: Tuple[int, int] = None,
                 gender_distribution: Tuple[float, float, float, float] = None,
                 organisation: Reference = None,
                 generate_ids: bool = False):
        self.resource_type = Patient
        self.n = n
        self.age_range = age_range
        self.gender_distribution = gender_distribution
        self._birthdate_range = None
        self.organisation = organisation
        self.generate_ids = generate_ids
        self.resources = None

    def generate(self, display: bool = False, references: bool = False):
        patients = self._generate(display=display)
        self.resources = patients

        if references and not self.generate_ids:
            raise ValueError("Cannot generate references without generating ids")
        elif references:
            return patients, self._generate_references()

        return patients

    def _generate(self, display: bool = False):
        patients = []
        names = self._generate_patient_names(self.n)
        for i in tqdm(range(self.n), desc=f"Generating {self.n} patients", disable=not display):
            patient = self._generate_patient_data(name=names[i])
            patients.append(patient)
        return patients

    def _generate_patient_data(self, name: Tuple[str, str]) -> Patient:

        first_name, last_name = name
        gender = random.choices(
            ["male", "female", "other", "unknown"],
            weights=self.gender_distribution if self.gender_distribution else [0.45, 0.45, 0.1, 0.0], k=1)[0]

        name = HumanName(**{"family": last_name, "given": [first_name]})

        birthdate = self._generate_birthdate()
        patient_dict = {
            "gender": gender,
            "name": [name],
            "birthDate": birthdate
        }
        if self.organisation:
            patient_dict["managingOrganization"] = self.organisation

        if self.generate_ids:
            patient_id = str(uuid4())
            patient_dict["id"] = patient_id

        return Patient(**patient_dict)

    @staticmethod
    def _generate_patient_names(n: int):

        p = Path(__file__).parent.joinpath("data").joinpath("first_names.txt")
        with open(p, "rb") as fnf:
            first_name_list = [fn.decode().strip().capitalize() for fn in fnf.readlines()]
        p = Path(__file__).parent.joinpath("data").joinpath("last_names.txt")
        with open(p, "rb") as lnf:
            last_name_list = [ln.decode().strip().capitalize() for ln in lnf.readlines()]

        first_names = random.choices(first_name_list, k=n)
        last_names = random.choices(last_name_list, k=n)
        names = list(zip(first_names, last_names))

        return names

    def _generate_birthdate(self):
        if not self._birthdate_range:
            if self.age_range:
                if isinstance(self.age_range[0], int):
                    # generate age range from 18-101 years old
                    now = pendulum.now()
                    youngest = pd.to_datetime((now - pendulum.duration(years=self.age_range[0])).to_date_string())
                    oldest = pd.to_datetime((now - pendulum.duration(years=self.age_range[1])).to_date_string())
                else:
                    raise ValueError(f"Unsupported type ({type(self.age_range[0])}) for generating patient ages."
                                     f"Only integers are supported.")
            else:
                # generate age range from 18-101 years old
                now = pendulum.now()
                youngest = pd.to_datetime((now - pendulum.duration(years=18)).to_date_string())
                oldest = pd.to_datetime((now - pendulum.duration(years=101)).to_date_string())

            self._birthdate_range = pd.date_range(oldest, youngest, freq="D").strftime('%Y-%m-%d').tolist()

        birthdate = random.choice(self._birthdate_range)
        return birthdate

    def _generate_references(self) -> List[Reference]:
        return [Reference(reference=f"Patient/{patient.id}") for patient in self.resources]

    def __repr__(self):
        return f"<PatientGenerator(n={self.n}, age_range={self.age_range}, " \
               f"gender_distribution={self.gender_distribution}>"



from typing import List
from requests import Response
from fhir.resources.resource import Resource
from fhir.resources.bundle import Bundle
from fhir.resources.reference import Reference
from pydantic import BaseModel


class CreateResponse:
    @staticmethod
    def _process_location_header(server_create_response: dict):
        # todo check for different fhir server types
        location_header: str = server_create_response.get("Location", server_create_response.get("location"))
        if not location_header:
            raise ValueError("Location field not found.")
        split_location = location_header.split("/")
        location = "/".join(split_location[:-2])
        version = int(split_location[-1])
        resource_id = split_location[-3]
        return resource_id, location, version


class ResourceCreateResponse(CreateResponse):
    location: str = None
    resource: Resource = None
    version: int = None
    resource_id: str = None
    reference: Reference = None

    def __init__(self, server_response_dict: dict, resource: Resource):
        self.resource = resource
        resource_id, location, version = self._process_location_header(server_response_dict)
        self.location = location
        self.resource_id = resource_id
        self.resource.id = resource_id
        self.reference = Reference(**{
            "reference": self.resource.get_resource_type() + "/" + resource_id
        })

    def __repr__(self):
        return f"<{self.__class__.__name__}(resource_id={self.resource_id}, location={self.location}," \
               f" version={self.version}, resource=...)>"


class BundleCreateResponse:
    create_responses: List[ResourceCreateResponse] = None

    def __init__(self, server_response: Response, bundle: Bundle):
        self.create_responses = []
        for i, entry in enumerate(server_response.json()["entry"]):
            resource = bundle.entry[i].resource
            create_response = ResourceCreateResponse(entry["response"], resource)
            self.create_responses.append(create_response)

    @property
    def resources(self):
        return [r.resource for r in self.create_responses]

    @property
    def references(self):
        return [r.reference for r in self.create_responses]

    def __repr__(self):
        return f"BundleCreateResponse(create_responses={self.create_responses[0]}...{self.create_responses[-1]}, " \
               f"num_resources={len(self.create_responses)})"


class UpdateResponse:
    def __init__(self, server_response: Response):
        pass


class ResourceSummary(BaseModel):
    resource: str
    count: int


class ServerSummary(BaseModel):
    name: str
    resources: List[ResourceSummary]

    @property
    def available_resources(self) -> List[ResourceSummary]:
        return [r for r in self.resources if r.count > 0]



import collections
import pathlib
from typing import Union, List, Dict, Optional
from enum import Enum

import pandas as pd
from fhir.resources.bundle import Bundle
from fhir.resources import FHIRAbstractModel
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from pydantic import BaseModel
from requests import Session, Response
import xmltodict

from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters
from fhir_kindling.serde import flatten_resources


class OutputFormats(Enum):
    """
    Enum for the output formats.
    """
    JSON = "json"
    XML = "xml"
    CSV = "csv"


class DataframeFormat(Enum):
    LIST = "list"
    SINGLE = "single"


class IncludedResources(BaseModel):
    """
    Container for resources included in the query via the _include/_revinclude parameters.
    """
    resource_type: str
    resources: Optional[List[FHIRAbstractModel]] = None


class ResponseStatusCodes(str, Enum):
    OK = 200
    CREATED = 201
    NOT_FOUND = 404
    BAD_REQUEST = 400
    INTERNAL_SERVER_ERROR = 500


class QueryResponse:
    """
    Response object for the results of a FHIR query executed against a FHIR server.
    """

    def __init__(self,
                 session: Session,
                 response: Response,
                 query_params: FHIRQueryParameters,
                 output_format: str = "json",
                 limit: int = None):

        self.session = session
        self.format = output_format
        self._limit = limit
        self.query_params = query_params
        self.resource = query_params.resource
        self._resources = None
        self._included_resources = {}
        self._bundle = None
        self.status_code: ResponseStatusCodes = None

        # parse the response after the rest of the setup is complete
        self.response: Union[str, dict] = self._process_server_response(response)

    @property
    def resources(self) -> List[FHIRResourceModel]:
        """
        List of primary resources returned by the server.

        Returns:
            List of FHIRResourceModel objects returned by the server.

        """
        if self.format == "xml":
            raise NotImplementedError("Resource parsing not supported for xml format")
        else:
            # return empty list if no resource matched the query
            if not self.response.get("entry"):
                return []
            if self._resources is None:
                self._extract_resources()
            return self._resources

    @property
    def included_resources(self) -> List[IncludedResources]:
        """
        Returns the list of resources included in the search.
        Returns:
            List of IncludeResources objects containing the resource type and the list of resources of this type
            included in the search

        """
        if self.format == "xml":
            raise NotImplementedError("Resource parsing not supported for xml format")
        else:
            # return empty list if no resource matched the query
            if not self.response.get("entry"):
                return []
            # raise error if there aren't any included resources
            if not self.query_params.include_parameters:
                raise ValueError("No included resources defined in query.")
            # parse the included resources if they don't exist
            if not self._included_resources:
                self._extract_resources()

            included = []
            for resource_type, resources in self._included_resources.items():
                included.append(IncludedResources(resource_type=resource_type, resources=resources))
            return included

    def save(self, file_path: Union[str, pathlib.Path], format: str = "json"):
        """
        Save the response to a file.
        Args:
            file_path: path to the file to save the response to.
            format: output format one of xml|json|(csv)

        Returns:
            None
        """

        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)

        format = OutputFormats(format)
        # check that only xml queries can be saved as xml
        if self.format == OutputFormats.XML:
            if not format == OutputFormats.XML:
                raise NotImplementedError("XML query results can only be saved as XML")

        self.format = format
        if self.format == OutputFormats.XML:
            with open(file_path, "w") as f:
                f.write(self.response)

        elif self.format == OutputFormats.JSON:
            with open(file_path, "w") as f:
                bundle = Bundle(**self.response)
                f.write(bundle.json(indent=2))
        elif self.format == OutputFormats.CSV:
            if self.query_params.include_parameters:
                for included_resources in self.included_resources:
                    df = flatten_resources(included_resources.resources)
                    included__resource_path = f"{file_path.parent}/{file_path.stem}_" \
                                              f"included_{included_resources.resource_type}.csv"
                    df.to_csv(included__resource_path, index=False)
            df = flatten_resources(self.resources)
            df.to_csv(file_path, index=False)

    def to_dfs(self, format: str = "list") -> Union[List[pd.DataFrame], pd.DataFrame]:
        """
        Serialize the response to a list of pandas dataframes
        Args:
            format: list of dataframe (one for each resource in the response) or single dataframe

        Returns:
            List of pandas dataframes or a single dataframe containing the resources in the query
        """

        df_format = DataframeFormat(format)

        if df_format == DataframeFormat.LIST:
            dfs = [flatten_resources(self.resources)]
            if self.query_params.include_parameters:
                for included_resources in self.included_resources:
                    dfs.append(flatten_resources(included_resources.resources))

            return dfs

        elif df_format == DataframeFormat.SINGLE:
            raise NotImplementedError("Single dataframe output format not implemented")

    def _extract_resources(self):
        """
        Parse the resources from the server response bundle. Split into included resources and resources that match the
        query exactly.
        Returns:

        """
        if not self._resources:
            self._resources = []
        for entry in Bundle(**self.response).entry:

            # add the directly queried resource to the resources list
            if entry.resource.resource_type == self.resource:
                self._resources.append(entry.resource)

            # process included resources
            elif entry.search.mode == "include":
                # get list of included resources based on type, if it does not return empty list
                included_resources = self._included_resources.get(entry.resource.resource_type, [])
                # update the list with the entry and update the included resources dict
                included_resources.append(entry.resource)
                self._included_resources[entry.resource.resource_type] = included_resources

    def _process_server_response(self, response: Response) -> Union[Dict, Bundle, str]:
        """
        Handle the initial response from the server and resolve pagination if necessary.
        Args:
            response: the initial server response

        Returns:
            the server response, with pagination resolved if necessary

        """

        # if the format is xml resolve pagination and return xml string response
        if self.format == "xml":
            return self._resolve_xml_pagination(response)
        # otherwise, resolve json pagination and process further according to selected outcome
        else:
            json_bundle = self._resolve_json_pagination(response)
            if self.format in ["json", "dict"]:
                return json_bundle
            elif self.format == "bundle":
                return Bundle(**json_bundle)

    def _resolve_xml_pagination(self, server_response: Response) -> str:

        # parse the xml response and extract the initial entries
        initial_response = xmltodict.parse(server_response.text)
        entries = initial_response["Bundle"].get("entry")

        # if there are no entries, return the initial response
        if not entries:
            self.status_code = ResponseStatusCodes.NOT_FOUND
            print(f"No resources match the query - query url: {self.query_params.to_query_string()}")
            return server_response.text
        else:
            self.status_code = ResponseStatusCodes.OK
        response = initial_response
        # resolve the pagination
        while True:
            next_page = False
            for link in response["Bundle"]["link"]:
                if isinstance(link, collections.OrderedDict):
                    relation_dict = dict(link["relation"])
                else:
                    break
                if relation_dict.get("@value") == "next":
                    # get url and extend with xml format
                    url = link["url"]["@value"]
                    url = url + "&_format=xml"  # todo check this
                    r = self.session.get(url)
                    r.raise_for_status()
                    response = xmltodict.parse(r.text)
                    added_entries = response["Bundle"]["entry"]
                    entries.extend(added_entries)
                    # Stop resolving the pagination when the limit is reached
                    if self._limit:
                        next_page = len(entries) < self._limit
                    else:
                        next_page = True

            if not next_page:
                print("All pages found")
                break
        # added the paginated resources to the initial response
        initial_response["Bundle"]["entry"] = entries[:self._limit] if self._limit else entries
        full_response_xml = xmltodict.unparse(initial_response, pretty=True)
        return full_response_xml

    def _resolve_json_pagination(self, server_response: Response):
        response = server_response.json()
        link = response.get("link", None)
        # If there is a link, get the next page otherwise return the response
        if not link:
            self.status_code = ResponseStatusCodes.OK
            return response
        else:
            entries = []
            initial_entry = response.get("entry", None)
            if not initial_entry:
                self.status_code = ResponseStatusCodes.NOT_FOUND
                print(f"No resources match the query - query url: {self.query_params.to_query_string()}")
                return response
            else:
                self.status_code = ResponseStatusCodes.OK
                entries.extend(response["entry"])
            # if the limit is reached, stop resolving the pagination
            if self._limit:
                if len(entries) >= self._limit:
                    response["entry"] = response["entry"][:self._limit]
                    return response
            # query the linked page and add the entries to the response
            while response.get("link", None):
                if self._limit and len(entries) >= self._limit:
                    print("Limit reached stopping pagination resolve")
                    break

                next_page = next((link for link in response["link"] if link.get("relation", None) == "next"), None)

                if next_page:
                    response = self.session.get(next_page["url"]).json()
                    entries.extend(response["entry"])
                else:
                    break

            response["entry"] = entries[:self._limit] if self._limit else entries
            return response

    def __repr__(self):
        return f"[{self.status_code}] QueryResponse<{self.resource}>:  total={len(self.resources)}\n" \
               f"Query string: {self.query_params.to_query_string()}\nparameters: {self.query_params}"



from typing import Union, List

from pydantic.fields import ModelField

from fhir.resources.resource import Resource
from fhir.resources import get_fhir_model_class
from fhir.resources.fhirtypes import ResourceType


def get_resource_fields(resource: Union[Resource, ResourceType, str]) -> List[ModelField]:
    """
    Get the fields of a resource.
    Args:
        resource:

    Returns:

    """
    # If the resource is a string, then we need to get the resource class
    if isinstance(resource, str):
        resource = get_fhir_model_class(resource)

    # Get the resource model fields
    fields = list(resource.__fields__.values())
    return fields


def valid_resource_name(resource_name: str) -> str:
    """

    Args:
        resource_name:

    Returns:

    """
    try:
        get_fhir_model_class(resource_name)
    except KeyError:
        raise ValueError(f"Invalid resource name: {resource_name}")
    return resource_name



from fhir.resources.patient import Patient

from fhir_kindling.util.resources import get_resource_fields


def test_get_resource_fields():
    fields = get_resource_fields(Patient)

    string_fields = get_resource_fields("Patient")

    assert fields == string_fields



import random
from typing import List, Union, Optional
from uuid import uuid4

from fhir.resources import get_fhir_model_class
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from fhir.resources.patient import Patient
from fhir.resources.reference import Reference
from pydantic import BaseModel, Field
from fhir.resources.fhirtypes import ReferenceType

from fhir_kindling import FhirServer
from fhir_kindling.generators.resource_generator import ResourceGenerator
from fhir_kindling.generators.patient import PatientGenerator
from fhir_kindling.util import get_resource_fields


class DataSetResourceGenerator(BaseModel):
    name: str
    generator: ResourceGenerator
    depends_on: Optional[Union[str, List[str]]] = None
    likelihood: float

    class Config:
        arbitrary_types_allowed = True


class GeneratedResources(BaseModel):
    resource_type: str
    resources: List[dict] = Field(default_factory=list)
    reference_key: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


class DataSet(BaseModel):
    name: str
    patients: List[Patient]
    resources: Optional[List[GeneratedResources]]

    class Config:
        arbitrary_types_allowed = True

    def upload(self, server: FhirServer) -> tuple:
        # first upload the patients for referential integrity
        patient_ids = [patient.id for patient in self.patients]
        patient_response = server.add_all(self.patients)
        # replace the initially generated reference with the server generated ones

        # iterate over the server generated references along with the initially generated ids
        for reference, patient_id in zip(patient_response.references, patient_ids):
            # iterate over all the generated resource objects
            for generated_resources in self.resources:
                # and their resources
                for resource in generated_resources.resources:
                    # replace the reference parameter with the server generated one
                    resource_reference = resource.get(generated_resources.reference_key)
                    if patient_id in resource_reference["reference"]:
                        resource[generated_resources.reference_key] = {"reference": reference.reference}

        # upload the resources
        generated_resources = []
        for generated_resources_object in self.resources:
            generated_resources.extend(generated_resources_object.resources)
        resource_response = server.add_all(generated_resources)

        return patient_response, resource_response


class DatasetGenerator:
    """
    Generates a dataset for a FHIR resource.
    """

    def __init__(self, resource: str = "Patient", n: int = None, name: str = None):

        self.name = name if name else str(uuid4())

        if resource == "Patient":
            self.resource = get_fhir_model_class(resource)
        else:
            raise NotImplementedError("DatasetGenerator for {} is not implemented".format(resource))
        self.n = n
        self.generators: List[DataSetResourceGenerator] = []
        self._references = None
        self._patients = None
        self._dataset: DataSet = None
        self._resource_types = set()
        self._reference_fields = {}

    def add_resource(self, resource_generator: ResourceGenerator, name: str = None,
                     depends_on: Union[str, List[str]] = None, likelihood: float = 1.0) -> 'DatasetGenerator':
        if not name:
            name = str(uuid4())
        self._resource_types.add(resource_generator.resource.get_resource_type())
        self.generators.append(
            DataSetResourceGenerator.construct(
                name=name,
                generator=resource_generator,
                depends_on=depends_on,
                likelihood=likelihood
            )
        )
        return self

    def generate(self, ids: bool = True) -> DataSet:
        """
        Generate a dataset of FHIR resources according to the given conditions

        Args:
            ids:

        Returns:

        """
        if self.generators:
            self._setup_dataset()
            self._patients, self._references = PatientGenerator(self.n, generate_ids=ids).generate(references=True)
            for reference in self._references:
                for generator in self.generators:
                    # todo resolve interdependencies
                    if random.random() > generator.likelihood:
                        continue
                    # generate an id to use in put request with the server
                    resource = generator.generator.generate(disable_validation=True, generate_ids=True)
                    self._add_reference_param(resource, reference)

                    # construct the resource object
                    resource_type = generator.generator.resource.get_resource_type()
                    resource = get_fhir_model_class(resource_type)(**resource.dict())

                    # store the generated resource
                    self._store_generated_resource(resource, resource_type=resource_type)

            # add the patient resources
            self._dataset.patients = self._patients
            # validate data set
            return DataSet(**self._dataset.dict())

        else:
            patient_generator = PatientGenerator(n=self.n if self.n else 100, generate_ids=ids)
            patients = patient_generator.generate()
            return DataSet(name=self.name, patients=patients)

    def explain(self):
        pass

    def _add_reference_param(self, resource: FHIRResourceModel, reference: Reference):
        # check if a reference field is present for the given resource type if not detect first required reference
        reference_field = self._reference_fields.get(
            resource.get_resource_type(),
            self._get_required_reference(resource)
        )
        resource.__setattr__(reference_field, reference)

    def _get_required_reference(self, resource: FHIRResourceModel) -> str:
        fields = get_resource_fields(resource)
        required_fields = []
        for field in fields:
            if field.required:
                # todo check this further
                # add the reference to the required reference fields
                if field.type_ == ReferenceType:
                    required_fields.append(field.name)
            if field.name in ["patient", "subject"]:
                required_fields.append(field.name)
        # if there is only one reference field return it

        if len(required_fields) == 1:
            reference = required_fields[0]
        # iterate over the reference fields and return the best match patient -> subject
        else:
            reference = None
            for field in set(required_fields):
                if field == "patient":
                    reference = field
                    break
                elif field == "subject":
                    reference = field
                else:
                    raise ValueError(f"No reference field found for {resource.get_resource_type()}")
        return reference

    def _store_generated_resource(self, resource: FHIRResourceModel, resource_type: str):
        # todo improve this
        for store in self._dataset.resources:
            if store.resource_type == resource_type:
                store.resources.append(resource.dict(exclude_none=True))

    def _setup_dataset(self):

        dataset = DataSet.construct(resources=[])
        if self.name:
            dataset.name = self.name
        for resource in self._resource_types:
            # find the reference field
            reference_field = self._get_required_reference(get_fhir_model_class(resource))
            self._reference_fields[resource] = reference_field
            # set up an initial empty resource store with
            dataset.resources.append(GeneratedResources(resource_type=resource, reference_key=reference_field))

        self._dataset = dataset

    def __repr__(self):
        return f"<{self.__class__.__name__}(name={self.name}, resource_types={self._resource_types}, n={self.n}," \
               f" generators={self.generators})>"



from abc import ABC

from pydantic import BaseModel, validator, root_validator
from enum import Enum
from typing import Union, Optional, List, Tuple

from fhir_kindling.util.resources import valid_resource_name


class QueryOperators(str, Enum):
    """
    Enumeration of query operators.
    """
    eq = "eq"
    ne = "ne"
    gt = "gt"
    lt = "lt"
    ge = "ge"
    le = "le"
    sa = "sa"
    eb = "eb"
    in_ = "in"
    not_in = "not_in"
    contains = "contains"
    regex = "regex"


"""
utility functions for parsing query parameters and values
"""


def check_url_param_primitives(value: str) -> Union[int, float, bool, str]:
    """
    Check if the string contains other primitives and return them before finally returning the string.

    Args:
        value: string of the value to check

    Returns:
        The value if it is a primitive type, otherwise return string.
    """
    if value.lower() == "true" or value.lower() == "false":
        return value.lower() == "true"
    # parsing order matters for ints and floats
    try:
        value = int(value)
        return value
    except ValueError:
        pass
    try:
        value = float(value)
        return value
    except ValueError:
        pass

    return value


def parse_parameter_value(url_value: str) -> Tuple[QueryOperators, Union[int, float, bool, str, list]]:
    """
    Parse a query parameter value for operators and value types

    Args:
        url_value: string snippet of the query url (after =)

    Returns: Tuple of the operator and the value

    """
    try:
        operator = QueryOperators(url_value[:2])
        value = url_value[2:]
        if operator == QueryOperators.ne and "," in value:
            value = [check_url_param_primitives(v) for v in value.split(",")]
            operator = QueryOperators.not_in
        else:
            value = check_url_param_primitives(value)

    except ValueError:
        value = url_value
        if "," in value:
            operator = QueryOperators.in_
            value = [check_url_param_primitives(v) for v in value.split(",")]
        else:
            operator = QueryOperators.eq
            value = check_url_param_primitives(value)
    return operator, value


class QueryParameter(BaseModel):
    """
    Base class for query parameters.
    """

    def to_url_param(self) -> str:
        """
        Convert the query parameter to a URL parameter.
        """
        raise NotImplementedError()

    @classmethod
    def from_url_param(cls, url_string: str):
        """
        Create a query parameter object from an url query snippet
        """
        raise NotImplementedError()


class QuerySearchParameter(QueryParameter, ABC):
    operator: QueryOperators
    value: Union[int, float, bool, str, list]

    @validator("value")
    def check_value(cls, v, values):
        if isinstance(v, list):
            if values["operator"] not in [QueryOperators.in_, QueryOperators.not_in]:
                raise ValueError(
                    "List values can only be used with the 'in' and 'not_in' operator."
                )
            for item in v:
                if not isinstance(item, (str, int, float, bool)):
                    raise ValueError(f"Invalid value type: {type(item)}")
        else:
            if values.get("operator") in [QueryOperators.in_, QueryOperators.not_in]:
                raise ValueError("The 'in' and 'not_in' operators can only be used with a list value.")

        return v

    def make_search_parameter_values(self) -> Tuple[str, str]:
        if self.operator in [QueryOperators.eq, QueryOperators.in_]:
            operator_prefix = ""
        else:
            operator_prefix = self.operator.value
        if isinstance(self.value, list):
            query_value = ",".join(self.value)
        elif isinstance(self.value, bool):
            query_value = str(self.value).lower()
        else:
            query_value = self.value

        return operator_prefix, query_value


class IncludeParameter(QueryParameter):
    """
    Class to represent include parameters in a fhir query
    """

    resource: str
    search_param: str
    target: Optional[str] = None
    reverse: bool = False
    iterate: bool = False

    def to_url_param(self) -> str:
        query_param = "_include" if not self.reverse else "_revinclude"

        iterate = ":iterate" if self.iterate else ""
        target = f"&{self.target}" if self.target else ""
        url_param = f"{query_param}{iterate}={self.resource}:{self.search_param}{target}"
        return url_param

    @classmethod
    def from_url_param(cls, url_string: str) -> "IncludeParameter":
        field, param = url_string.split("=")
        split_field = field.split(":")
        if len(split_field) == 2:
            reverse = split_field[0] == "_revinclude"
            iterate = split_field[1] == "iterate"
            if not iterate:
                raise ValueError(
                    f"Invalid include iterate parameter in: {url_string}\n\t {field} must contain ':iterate'")
        else:
            reverse = split_field[0] == "_revinclude"
            iterate = False
        resource, search_param = param.split(":")

        param_split = search_param.split("&")
        if len(param_split) == 2:
            target = param_split[1]
            search_param = param_split[0]
        else:
            target = None
            search_param = param_split[0]

        return cls(
            resource=resource,
            search_param=search_param,
            target=target,
            iterate=iterate,
            reverse=reverse
        )


class ReverseChainParameter(QuerySearchParameter):
    """
    Class to represent reverse chain parameters in a fhir query for querying resources based on properties of other
    resources that refer to them.
    """
    resource: str
    reference_param: str
    search_param: str
    operator: QueryOperators
    value: Union[int, float, list, bool, str]

    _normalize_resource = validator("resource", allow_reuse=True)(valid_resource_name)

    class Config:
        smart_union = True

    def to_url_param(self) -> str:
        operator_prefix, query_value = self.make_search_parameter_values()
        url_param = f"_has:{self.resource}:{self.reference_param}:{self.search_param}={operator_prefix}{query_value}"

        return url_param

    @classmethod
    def from_url_param(cls, url_string: str) -> "ReverseChainParameter":
        chained_field, param = url_string.split("=")
        _, resource, reference_param, search_param = chained_field.split(":")
        operator, value = parse_parameter_value(url_value=param)
        return cls(
            resource=resource,
            reference_param=reference_param,
            search_param=search_param,
            operator=operator,
            value=value
        )


class FieldParameter(QuerySearchParameter):
    """
    The Field Query Parameter class.
    """
    field: str

    class Config:
        smart_union = True

    def to_url_param(self) -> str:
        operator_prefix, query_value = self.make_search_parameter_values()

        url_param = f"{self.field}={operator_prefix}{query_value}"

        return url_param

    @classmethod
    def from_url_param(cls, url_string: str) -> "FieldParameter":
        field, param = url_string.split("=")
        operator, value = parse_parameter_value(url_value=param)
        return FieldParameter(
            field=field,
            operator=operator,
            value=value
        )


class FHIRQueryParameters(BaseModel):
    """
    Collection of query parameters for a fhir query.
    """
    resource: str
    resource_parameters: Optional[List[FieldParameter]] = None
    include_parameters: Optional[List[IncludeParameter]] = None
    has_parameters: Optional[List[ReverseChainParameter]] = None

    _normalize_resource = validator("resource", allow_reuse=True)(valid_resource_name)

    @root_validator
    def validate_parameters(cls, values):
        # todo
        return values

    def to_query_string(self) -> str:
        """
        Converts the parameters to a query string that can be used with a fhir server's REST API
        Returns:
        """
        query_string = f"/{self.resource}?"
        if self.resource_parameters:
            resource_url_params = "&".join([param.to_url_param() for param in self.resource_parameters])
            query_string += resource_url_params
        # include parameters
        if self.include_parameters:
            if self.resource_parameters:
                query_string += "&"
            include_url_params = "&".join([param.to_url_param() for param in self.include_parameters])
            query_string += include_url_params
        # reverse chain parameters
        if self.has_parameters:
            if self.resource_parameters or self.include_parameters:
                query_string += "&"
            has_url_params = "&".join([param.to_url_param() for param in self.has_parameters])
            query_string += has_url_params

        return query_string

    @classmethod
    def from_query_string(cls, query_string: str) -> "FHIRQueryParameters":

        # split resource and query parameters
        resource, query = query_string.split("?")

        # clean up and validate resource
        if resource[0] == "/":
            resource = resource[1:]
        resource = valid_resource_name(resource)

        resource_parameters = None
        include_parameters = None
        has_parameters = None
        if query:
            # parse query parameters
            query_params = query.split("&")
            resource_parameters = []
            include_parameters = []
            has_parameters = []
            for param in query_params:
                # start with the special keywords and finally attempt to parse as resource query
                if param.startswith("_has"):
                    has_param = ReverseChainParameter.from_url_param(param)
                    has_parameters.append(has_param)
                elif param.startswith("_include") or param.startswith("_revinclude"):
                    include_param = IncludeParameter.from_url_param(param)
                    include_parameters.append(include_param)
                else:
                    resource_param = FieldParameter.from_url_param(param)
                    resource_parameters.append(resource_param)

        return cls(
            resource=resource,
            resource_parameters=resource_parameters,
            include_parameters=include_parameters,
            has_parameters=has_parameters
        )



from typing import List, Union, Optional, Any
from fhir.resources import get_fhir_model_class
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from uuid import uuid4
from fhir.resources.resource import Resource
from pydantic import BaseModel

from fhir_kindling.generators.field_generator import FieldGenerator


class FieldValue(BaseModel):
    field: str
    value: Union[Any, List[Any]]


class GeneratorParameters(BaseModel):
    count: Optional[int] = None
    field_generators: Optional[List[FieldGenerator]] = None
    field_values: Optional[List[FieldValue]] = None

    # todo count and list in field values of same length


class ResourceGenerator:

    def __init__(self, resource: Union[str, Any], n: int = None, field_values: dict = None,
                 disable_validation: bool = False,
                 generator_parameters: GeneratorParameters = None):
        if not isinstance(resource, str):
            try:
                resource_type = resource.get_resource_type()
                self.resource = get_fhir_model_class(resource_type)
            except Exception:
                raise ValueError(f"Resource must be a string or a FHIRResourceModel, got {type(resource)}")
        else:
            self.resource = get_fhir_model_class(resource)
        self.params = generator_parameters
        self.field_values = field_values
        if self.field_values and not disable_validation:
            self._check_required_fields()
        self.disable_validation = disable_validation
        self.n = n
        self._value_iterators = {}
        # list to store the field names of all fields being generated
        self._generated_fields = set()

    def required_fields(self) -> List[str]:
        required_fields = []
        for field_name, field in self.resource.__fields__.items():
            if field.required:
                required_fields.append(field_name)
        return required_fields

    def fields(self):
        return self.resource.__fields__

    def generate(self, disable_validation: bool = False, generate_ids: bool = False) -> Union[Resource, List[Resource]]:

        self.disable_validation = disable_validation
        # if field values are given parse them into parameters
        if self.n and not self.params:
            self.params = GeneratorParameters(count=self.n)

        if self.field_values and self.n:
            self._parse_field_values()
        if not self.disable_validation:
            self._validate_params()
        resources = self._generate_resources(generate_ids)
        return resources

    def _generate_resources(self, generate_ids: bool):
        resources = []
        if self.params.count:
            for i in range(self.params.count):
                resource = self._generate_resource(generate_ids)
                resources.append(resource)
            return resources
        else:
            return self._generate_resource(generate_ids)

    def _generate_resource(self, generate_id: bool) -> FHIRResourceModel:
        # construct a resource object to hold the generated fields
        resource = self.resource.construct()
        # disable assignment validation
        resource.Config.validate_assignment = False
        if self.params.field_values:
            for field_value in self.params.field_values:
                # update resource with field value
                self._update_with_field_value(resource, field_value)

        if self.params.field_generators:
            for generator in self.params.field_generators:
                # update resource with generated field value
                self._update_with_field_generator(resource, generator)
        if generate_id:
            resource.id = str(uuid4())

        # validate resource when validation is enabled
        if not self.disable_validation:
            resource = self.resource(**resource.dict(exclude_none=True))

        return resource

    def _update_with_field_value(self, resource: FHIRResourceModel, field_value: FieldValue):
        if isinstance(field_value.value, list):
            iterator = self._value_iterators.get(field_value.field)
            if not iterator:
                iterator = iter(field_value.value)
                self._value_iterators[field_value.field] = iterator
            resource.__setattr__(field_value.field, next(iterator))
        else:
            resource.__setattr__(field_value.field, field_value.value)

    def _update_with_field_generator(self, resource: FHIRResourceModel, field_generator: FieldGenerator):

        value = field_generator.generate()
        resource.__setattr__(field_generator.field, value)

    def _check_required_fields(self):
        """
        Check if all required fields are being generated by the field generators or are given as field values

        Returns:

        """
        required_fields_set = set(self.required_fields())
        if not required_fields_set.issubset(self._generated_fields):
            raise ValueError(f"Required fields {required_fields_set - self._generated_fields} not generated, "
                             f"generated fields: {self._generated_fields}")

    def _parse_field_values(self):

        params = GeneratorParameters(
            count=self.n
        )

        # todo: add support for field_values as a list of dicts

        self.params = params

    def _validate_params(self):
        # validate field values
        if self.params.field_values:
            self._validate_field_values()

        # validate field generators
        if self.params.field_generators:
            self._validate_field_generators()

        # check that the required fields are being generated
        self._check_required_fields()

    def _validate_field_values(self):
        field_values = self.params.field_values
        resource_count = self.params.count

        for field_value in field_values:
            # check for duplicates in generated fields
            if field_value.field in self._generated_fields:
                raise ValueError(f"Field value {field_value.value} is already generated")
            self._generated_fields.add(field_value.field)

            # check that the list length matches the resource count
            if isinstance(field_value.value, list):
                if len(field_value.value) != resource_count:
                    raise ValueError(f"Field value list length does not match resource count: {field_value.field}"
                                     f"Items in field value list: {len(field_value.value)},"
                                     f" resource count: {resource_count}")
            # todo check that the type of the item fits into the selected field

    def _validate_field_generators(self):
        generators = self.params.field_generators
        for resource_generator in generators:
            # check for duplicates in generated fields
            if resource_generator.field in self._generated_fields:
                raise ValueError(f"Field generator {resource_generator.field} is already generated")
            self._generated_fields.add(resource_generator.field)
            # todo validate generator values

    def __repr__(self):
        return f"<{self.__class__.__name__}(n={self.n}, resource={self.resource}>"
